# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from metrax.nnx import nnx_metrics

AUCPR = nnx_metrics.AUCPR
AUCROC = nnx_metrics.AUCROC
Accuracy = nnx_metrics.Accuracy
Average = nnx_metrics.Average
AveragePrecisionAtK = nnx_metrics.AveragePrecisionAtK
BLEU = nnx_metrics.BLEU
DCGAtK = nnx_metrics.DCGAtK
Dice = nnx_metrics.Dice
FBetaScore = nnx_metrics.FBetaScore
IoU = nnx_metrics.IoU
MAE = nnx_metrics.MAE
MRR = nnx_metrics.MRR
MSE = nnx_metrics.MSE
NDCGAtK = nnx_metrics.NDCGAtK
Perplexity = nnx_metrics.Perplexity
Precision = nnx_metrics.Precision
PrecisionAtK = nnx_metrics.PrecisionAtK
PSNR = nnx_metrics.PSNR
RMSE = nnx_metrics.RMSE
RSQUARED = nnx_metrics.RSQUARED
Recall = nnx_metrics.Recall
RecallAtK = nnx_metrics.RecallAtK
RougeL = nnx_metrics.RougeL
RougeN = nnx_metrics.RougeN
SNR = nnx_metrics.SNR
SSIM = nnx_metrics.SSIM
WER = nnx_metrics.WER


__all__ = [
    "AUCPR",
    "AUCROC",
    "Average",
    "AveragePrecisionAtK",
    "BLEU",
    "DCGAtK",
    "Dice",
    "FBetaScore",
    "IoU",
    "MRR",
    "MAE",
    "MSE",
    "NDCGAtK",
    "Perplexity",
    "Precision",
    "PrecisionAtK",
    "PSNR",
    "RMSE",
    "RSQUARED",
    "Recall",
    "RecallAtK",
    "RougeL",
    "RougeN",
    "SNR",
    "SSIM",
    "WER",
]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of different metrics for ranking models."""

import abc
import flax
import jax
import jax.numpy as jnp
from metrax import base


@flax.struct.dataclass
class DCGAtK(base.Average):
  r"""Computes Discounted Cumulative Gain at k metric.

  DCG tells how good a list of search results or recommendations is, based on
  the relevance of the items and their positions in the list.

  This implementation calculates :math:`DCG@k` based on the following formula:

  .. math::
      DCG@K(y, s) = \sum_{i=1}^{K} \text{gain}(y_i) \times
      \text{rank_discount}(\text{rank}(s_i))

  where

    - :math:`y_i` is the relevance label from the labels,
    - :math:`s_i` is its score from the prediction,
    - :math:`\text{rank}(s_i)` is the 1-based rank of item :math:`i`.
    - :math:`\text{gain}(y_i) = 2^{y_i} - 1`.
    - :math:`\text{rank_discount}(\text{rank}(s_i)) = \frac{1}{\log_2(\text{rank}(s_i) + 1)}`.

  We get the final formula:

  .. math::
      DCG@K(y, s) = \sum_{i=1}^{K} \frac{2^{y_i} - 1}{\log_2(\text{rank}(s_i) +
      1)}
  """

  @classmethod
  def _calculate_dcg_at_ks(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> jax.Array:
    """Computes DCG@k for each example and for each k, using 'exp2' gain.

    This function is JIT-compiled. The gain calculation is fixed to 'exp2'.
    It uses jax.vmap to compute DCG for multiple k values efficiently.

    Args:
      predictions: A floating point 2D array (batch_size, vocab_size)
        representing prediction scores. Higher scores mean higher rank.
      labels: A 2D array (batch_size, vocab_size) of graded relevance scores.
      ks: A 1D array of integers representing the k values for which DCG is
        computed (e.g., jnp.array([1, 5, 10])). Shape: (|ks|,).

    Returns:
      A 2D array (batch_size, |ks|) containing DCG@k values.
    """
    gains = jnp.power(2.0, labels.astype(jnp.float32)) - 1.0
    score_ranks = jnp.argsort(jnp.argsort(-predictions, axis=1), axis=1) + 1
    score_rank_discounts = 1.0 / jnp.log2(score_ranks.astype(jnp.float32) + 1.0)
    item_contributions = gains * score_rank_discounts

    def _compute_dcg_at_k(k, current_item_contributions, current_score_ranks):
      """Computes DCG for a single k value across all examples in a batch.

      Args:
        k: A scalar array representing the single 'k' (top-k) value for which
          DCG is to be computed.
        current_item_contributions: A 2D array containing the pre-calculated
          contribution (gain * discount) for each item in each example of the
          batch. The shape should be (batch_size, vocab_size).
        current_score_ranks: A 2D array containing the 1-based rank for each
          item in each example of the batch. The shape should be (batch_size,
          vocab_size).

      Returns:
        A 1D array containing the DCG@k for each example in the batch.
        The shape should be (batch_size, ).
      """
      mask_for_k = current_score_ranks <= k
      dcg_at_k = jnp.sum(current_item_contributions * mask_for_k, axis=1)
      return dcg_at_k

    dcg_at_ks = jax.vmap(
        _compute_dcg_at_k,
        in_axes=(0, None, None),
        out_axes=1,  # Place the mapped axis(from ks) as the second axis
    )(ks, item_contributions, score_ranks)

    return dcg_at_ks

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> 'DCGAtK':
    """Creates a DCGAtK metric instance from model output."""
    dcg_at_ks = cls._calculate_dcg_at_ks(predictions, labels, ks)
    num_examples = jnp.array(labels.shape[0], dtype=jnp.float32)
    return cls(
        total=jnp.sum(dcg_at_ks, axis=0),
        count=num_examples,
    )


@flax.struct.dataclass
class NDCGAtK(DCGAtK):
  r"""Computes Normalized Discounted Cumulative Gain at k (NDCG@k) metrics.

  NDCG@k normalizes DCG@k by the Ideal DCG@k (IDCG@k), which is the DCG@k
  score of a perfectly ranked list (items sorted by their true relevance).

  This implementation calculates :math:`NDCG@k` based on the following formula:

  .. math::
      NDCG@k = \frac{DCG@k}{IDCG@k}

  where

    - If :math:`IDCG@k` is 0, then :math:`NDCG@k` is defined as 0.
    - The :math:`DCG@k` calculation uses :math:`exp2` gain
      (:math:`2^{\text{relevance}} - 1`) and standard logarithmic discount
      (:math:`\frac{1}{\log_2(\text{rank} + 1)}`).
  """

  @classmethod
  def _calculate_dcg_at_ks(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> jax.Array:
    """Computes NDCG@k for each example and for each k.

    Args:
      predictions: A floating point 2D array (batch_size, vocab_size)
        representing prediction scores.
      labels: A 2D array (batch_size, vocab_size) of graded relevance scores.
      ks: A 1D array of integers representing the k values. The shape should be
        (|ks|, ).

    Returns:
      A 2D array (batch_size, |ks|) containing NDCG@k values.
    """
    actual_dcg_at_ks = super()._calculate_dcg_at_ks(predictions, labels, ks)
    ideal_dcg_at_ks = super()._calculate_dcg_at_ks(
        predictions=labels,  # Use labels to determine ideal DCG.
        labels=labels,
        ks=ks,
    )
    ndcg_at_ks = base.divide_no_nan(actual_dcg_at_ks, ideal_dcg_at_ks)
    # Ensure NDCG is not greater than 1.
    ndcg_at_ks = jnp.minimum(ndcg_at_ks, 1.0)
    return ndcg_at_ks


@flax.struct.dataclass
class AveragePrecisionAtK(base.Average):
  r"""Computes AP@k (average precision at k) metrics.

  Average precision at k (AP@k) is a metric used to evaluate the performance of
  ranking models. It measures the sum of precision at k where the item at
  the kth rank is relevant, divided by the total number of relevant items.

  Given the top :math:`K` recommendations, AP@K is calculated as:

  .. math::
      AP@K = \frac{1}{r} \sum_{k=1}^{K} Precision@k * rel(k) \\
      rel(k) =
        \begin{cases}
          1 & \text{if the item at rank } k \text{ is relevant} \\
          0 & \text{otherwise}
        \end{cases}
  """

  @classmethod
  def average_precision_at_ks(
      cls, predictions: jax.Array, labels: jax.Array, ks: jax.Array
  ):
    """Computes AP@k (average precision at k) metrics for each of k in ks.

    Args:
      predictions: A floating point 2D vector representing the prediction
        generated from the model. The shape should be (batch_size, vocab_size).
      labels: A multi-hot encoding of the true label. The shape should be
        (batch_size, vocab_size).
      ks: A 1D vector of integers representing the k's to compute the MAP@k
        metrics. The shape should be (|ks|).

    Returns:
      Rank-2 tensor of shape (batch, |ks|) containing AP@k metrics.
    """
    indices_by_rank = jnp.argsort(-predictions, axis=1)
    labels = jnp.array(labels >= 1, dtype=jnp.float32)
    total_relevant = labels.sum(axis=1)

    def compute_ap_at_k_single(relevant_labels, total_relevant, ks):
      cumulative_precision = jnp.where(
          relevant_labels,
          base.divide_no_nan(
              jnp.cumsum(relevant_labels),
              jnp.arange(1, len(relevant_labels) + 1),
          ),
          0,
      )
      return jnp.array([
          base.divide_no_nan(
              jnp.sum(
                  jnp.where(
                      jnp.arange(cumulative_precision.shape[0]) < k,
                      cumulative_precision,
                      0.0,
                  )
              ),
              total_relevant,
          )
          for k in ks
      ])

    vmap_compute_ap_at_k = jax.vmap(
        compute_ap_at_k_single, in_axes=(0, 0, None), out_axes=0
    )

    ap_at_ks = vmap_compute_ap_at_k(
        jnp.take_along_axis(labels, indices_by_rank, axis=1), total_relevant, ks
    )
    return ap_at_ks

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> 'AveragePrecisionAtK':
    """Updates the metric.

    Args:
      predictions: A floating point 2D vector representing the prediction
        generated from the model. The shape should be (batch_size, vocab_size).
      labels: A multi-hot encoding of the true label. The shape should be
        (batch_size, vocab_size).
      ks: A 1D vector of integers representing the k's to compute the MAP@k
        metrics. The shape should be (|ks|).

    Returns:
      The AveragePrecisionAtK metric. The shape should be (|ks|).

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    ap_at_ks = cls.average_precision_at_ks(predictions, labels, ks)
    num_examples = jnp.array(labels.shape[0], dtype=jnp.float32)
    return cls(
        total=ap_at_ks.sum(axis=0),
        count=num_examples,
    )


@flax.struct.dataclass
class MRR(base.Average):
  r"""Computes Mean Reciprocal Rank (MRR), supporting MRR@k for multiple k values.

  MRR is the average of the reciprocal ranks of the first relevant item
  for a set of queries.

  The mean reciprocal rank for a group of queries :math:`q` in :math:`Q` is defined as follows:

  .. math::
      MRR = \frac{1}{|Q|} \sum_{q \in Q} RR_q

  Where :math:`RR_q` is the reciprocal rank for query :math:`q`, defined as:

  .. math::
      RR_q =
        \begin{cases}
          \frac{1}{\text{rank}} & \text{if a revelant item is found} \\
          0 & \text{if no relevant item is found.}
        \end{cases}

  This implementation assumes binary relevance labels (1 for relevant, 0 for not
  relevant).
  """

  @classmethod
  def _calculate_rr_at_ks(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> jax.Array:
    """Calculates the Reciprocal Rank@k for each query in a batch, for each k.

    Args:
      predictions: A 2D array of prediction scores. Higher scores indicate
        higher rank. The shape should be (batch_size, vocab_size).
      labels: A 2D array of binary relevance labels (0 or 1). The shape should
        be (batch_size, vocab_size).
      ks: A 1D array of integers representing the k cutoffs. The shape should be
        (|ks|,).

    Returns:
      A 2D array containing the Reciprocal Rank@k for each query and each k.
      Shape: (batch_size, |ks|).
    """
    indices_by_rank = jnp.argsort(-predictions, axis=1)
    labels_by_rank = jnp.take_along_axis(labels, indices_by_rank, axis=1)
    first_relevant_indices = jnp.argmax(labels_by_rank, axis=1)
    first_relevant_ranks = first_relevant_indices + 1
    # Check if a relevant item exists for each query.
    reciprocal_rank = jnp.where(
        jnp.any(labels_by_rank > 0, axis=1),
        1.0 / first_relevant_ranks.astype(jnp.float32),
        0.0,
    )

    def _compute_rr_at_k(k, reciprocal_rank, first_relevant_rank):
      rr_at_k = jnp.where(first_relevant_rank <= k, reciprocal_rank, 0.0)
      return rr_at_k

    rr_at_ks = jax.vmap(_compute_rr_at_k, in_axes=(0, None, None), out_axes=1)(
        ks, reciprocal_rank, first_relevant_ranks
    )

    return rr_at_ks

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> 'MRR':
    """Creates an MRR metric instance from model output, calculating MRR@k for each k.

    Args:
      predictions: A 2D array of prediction scores. Higher scores indicate
        higher rank. The shape should be (batch_size, vocab_size).
      labels: A 2D array of binary relevance labels (0 or 1). The shape should
        be (batch_size, vocab_size).
      ks: A 1D array of integers representing the k cutoffs. The shape should be
        (|ks|, ).

    Returns:
      An MRR metric object. The 'total' field will be an array of shape
      (|ks|, ).
    """
    rr_at_ks = cls._calculate_rr_at_ks(predictions, labels, ks)
    num_queries = jnp.array(labels.shape[0], dtype=jnp.float32)
    return cls(
        total=rr_at_ks.sum(axis=0),
        count=num_queries,
    )


@flax.struct.dataclass
class TopKRankingMetric(base.Average, abc.ABC):
  """Abstract base class for Top-K ranking metrics like Precision@k and Recall@k.

  This class provides common functionality for calculating metrics that evaluate
  the quality of the top k items in a ranked list. Subclasses must implement
  the `_calculate_metric_at_ks` method to define the specific metric
  computation (e.g., precision, recall).

  The `from_model_output` method is a factory method that computes the metric
  values for a batch of predictions and labels, and aggregates them.
  """

  @staticmethod
  def _get_relevant_at_k(
      predictions: jax.Array, labels: jax.Array, ks: jax.Array
  ) -> jax.Array:
    """Computes the number of relevant items at each k.

    This static method processes predictions and labels to determine the
    number of relevant items at specified k-values.

    Args:
      predictions: A floating point 2D array representing the prediction scores
        from the model. Higher scores indicate higher relevance. The shape
        should be (batch_size, vocab_size).
      labels: A multi-hot encoding (0 or 1) of the true labels. The shape should
        be (batch_size, vocab_size).
      ks: A 1D array of integers representing the k's to compute the P@k
        metrics. The shape should be (|ks|).

    Returns:
      relevant_at_k: A 2D array of shape (batch_size, |ks|). Each element [i, j]
      is the number of relevant items among the top ks[j] recommendations for
      the i-th example in the batch.
    """
    labels = jnp.array(labels >= 1, dtype=jnp.float32)
    indices_by_rank = jnp.argsort(-predictions, axis=1)
    labels_by_rank = jnp.take_along_axis(labels, indices_by_rank, axis=1)
    relevant_by_rank = jnp.cumsum(labels_by_rank, axis=1)

    vocab_size = predictions.shape[1]
    k_indices = jnp.minimum(ks - 1, vocab_size - 1)
    relevant_at_k = relevant_by_rank[:, k_indices]

    return relevant_at_k

  @classmethod
  @abc.abstractmethod
  def _calculate_metric_at_ks(
      cls, predictions: jax.Array, labels: jax.Array, ks: jax.Array
  ) -> jax.Array:
    """Computes the specific metric (e.g., P@k, R@k) values per example for each k.

    This method must be implemented by concrete subclasses (e.g., PrecisionAtK,
    RecallAtK) to define the actual calculation of the metric based on
    predictions, labels, and k-values.

    Args:
      predictions: A floating point 2D array representing the prediction scores
        from the model.
      labels: A multi-hot encoding of the true labels.
      ks: A 1D array of integers representing the k's.

    Returns:
      A rank-2 array of shape (batch_size, |ks|) containing the metric
      values for each example in the batch and each specified k.
    """
    raise NotImplementedError('Subclasses must implement this method.')

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      ks: jax.Array,
  ) -> 'TopKRankingMetric':
    """Creates a metric instance from model output.

    This class method computes the specific ranking metric (defined by the
    subclass's implementation of `_calculate_metric_at_ks`) for each example
    in the batch. It then aggregates these values (sum of metric values and
    count of examples) into a metric object. This object can later be used
    to compute the mean metric value (e.g., Mean Precision@k) by calling
    its `.compute()` method (inherited from `base.Average`).

    Args:
      predictions: A floating point 2D array representing the prediction scores
        from the model. The shape should be (batch_size, vocab_size).
      labels: A multi-hot encoding (0 or 1, or counts) of the true labels. The
        shape should be (batch_size, vocab_size).
      ks: A 1D array of integers representing the k's to compute the metrics.
        The shape should be (|ks|).

    Returns:
      An instance of the calling class (e.g., PrecisionAtK, RecallAtK)
      with `total` and `count` fields populated. The `total` field will
      have shape (|ks|), representing the sum of metric values for each k
      across the batch, and `count` will be a scalar representing the
      number of examples in the batch.
    """
    metric_at_ks = cls._calculate_metric_at_ks(predictions, labels, ks)
    num_examples = jnp.array(labels.shape[0], dtype=jnp.float32)
    return cls(
        total=metric_at_ks.sum(axis=0),
        count=num_examples,
    )


@flax.struct.dataclass
class PrecisionAtK(TopKRankingMetric):
  r"""Computes P@k (precision at k) metrics.

  Precision at k (P@k) is a metric that measures the proportion of
  relevant items found in the top k recommendations. It answers the question:
  "Out of the K items recommended, how many are actually relevant?"

  Given the top :math:`K` recommendations, P@K is calculated as:

  .. math::
      Precision@K = \frac{\text{Number of relevant items in top K}}{K}
  """

  @classmethod
  def _calculate_metric_at_ks(
      cls, predictions: jax.Array, labels: jax.Array, ks: jax.Array
  ) -> jax.Array:
    """Computes P@k (precision at k) metrics for each of k in ks for each example.

    This method implements the core logic for calculating Precision@k.
    It utilizes the `_get_relevant_at_k` helper from the base
    class to get the number of relevant items at each k, and then divides
    by k (clamped by vocabulary size) to get the precision.

    Args:
      predictions: A floating point 2D array representing the prediction scores
        from the model. The shape should be (batch_size, vocab_size).
      labels: A multi-hot encoding (0 or 1, or counts) of the true labels. The
        shape should be (batch_size, vocab_size).
      ks: A 1D array of integers representing the k's to compute the P@k
        metrics. The shape should be (|ks|).

    Returns:
      A rank-2 array of shape (batch_size, |ks|) containing P@k metrics
      for each example and each k.
    """
    relevant_at_k = cls._get_relevant_at_k(predictions, labels, ks)
    vocab_size = labels.shape[1]
    denominator_p_at_k = jnp.minimum(ks.astype(jnp.float32), vocab_size)
    return base.divide_no_nan(relevant_at_k, denominator_p_at_k[jnp.newaxis, :])


@flax.struct.dataclass
class RecallAtK(TopKRankingMetric):
  r"""Computes R@k (recall at k) metrics.

  Recall at k (R@k) is a metric that measures the proportion of
  relevant items that are found in the top k recommendations, out of the
  total number of relevant items for a given user/query. It answers the
  question:
  "Out of all the items that are truly relevant, how many did we find in the top
  K?"

  Given the top :math:`K` recommendations, R@K is calculated as:

  .. math::
      Recall@K = \frac{\text{Number of relevant items in top K}}{\text{Total
      number of relevant items}}
  """

  @classmethod
  def _calculate_metric_at_ks(
      cls, predictions: jax.Array, labels: jax.Array, ks: jax.Array
  ) -> jax.Array:
    """Computes R@k (recall at k) metrics for each of k in ks for each example.

    This method implements the core logic for calculating Recall@k.
    It utilizes the `_get_relevant_at_k` helper from the base
    class to get the number of relevant items at each k and the binarized
    labels.
    The number of relevant items at k is then divided by the total number of
    relevant items for that example to get the recall.

    Args:
      predictions: A floating point 2D array representing the prediction scores
        from the model. The shape should be (batch_size, vocab_size).
      labels: A multi-hot encoding (0 or 1, or counts) of the true labels. The
        shape should be (batch_size, vocab_size).
      ks: A 1D array of integers representing the k's to compute the R@k
        metrics. The shape should be (|ks|).

    Returns:
      A rank-2 array of shape (batch_size, |ks|) containing R@k metrics
      for each example and each k.
    """
    relevant_at_k = cls._get_relevant_at_k(predictions, labels, ks)
    total_relevant = jnp.sum(jnp.array(labels >= 1, dtype=jnp.float32), axis=1)
    return base.divide_no_nan(relevant_at_k, total_relevant[:, jnp.newaxis])



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metrax LoggingBackend protocol."""

from typing import Protocol
import numpy as np


class LoggingBackend(Protocol):
  """Defines the interface for a pluggable logging backend."""

  def log_scalar(self, event: str, value: float | np.ndarray, **kwargs):
    """Logs a scalar value. Must match jax.monitoring listener signature."""
    ...

  def close(self):
    """Closes the logger and flushes any pending data."""
    ...



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of different metrics for audio models."""

import flax
import jax
import jax.numpy as jnp
from metrax import base


@flax.struct.dataclass
class SNR(base.Average):
  r"""SNR (Signal-to-Noise Ratio) Metric for audio.

  This class calculates the Signal-to-Noise Ratio (SNR) in decibels (dB)
  between a predicted audio signal and a ground truth audio signal,
  and averages it over a dataset.

  The SNR is defined as:

  .. math::

      SNR_{dB} = 10 \cdot \log_{10} \left( \frac{P_{signal}}{P_{noise}}
      \right)

  Where:
    - :math:`P_{signal}` is the power of the ground truth signal (`target`).
      By default (`zero_mean=False`), this is the mean of the squared `target`
      values.
      If `zero_mean=True`, it's the variance of the `target` values.
    - :math:`P_{noise}` is the power of the noise component, which is defined as
      the difference between the `target` and `preds` (`target - preds`).
      By default (`zero_mean=False`), this is the mean of the squared noise
      values.
      If `zero_mean=True`, it's the variance of the noise values.
  """

  @staticmethod
  def _calculate_snr(
      preds: jax.Array,
      target: jax.Array,
      zero_mean: bool = False,
  ) -> jax.Array:
    """Computes SNR (Signal-to-Noise Ratio) values for a batch of audio signals.

    Args:
        preds: The estimated or predicted audio signal. JAX Array.
        target: The ground truth audio signal. JAX Array.
        zero_mean: If True, subtracts the mean from the signal and noise before
          calculating their respective powers. Defaults to False.

    Returns:
        A 1D JAX array representing the SNR in decibels (dB) for each example
        in the batch.
    """
    if preds.shape != target.shape:
      raise ValueError(
          f'Input signals must have the same shape, but got {preds.shape} and'
          f' {target.shape}'
      )

    target_processed, preds_processed = jax.lax.cond(
        zero_mean,
        lambda t, p: (
            t - jnp.mean(t, axis=-1, keepdims=True),
            p - jnp.mean(p, axis=-1, keepdims=True),
        ),
        lambda t, p: (t, p),
        target,
        preds,
    )
    noise = target_processed - preds_processed
    eps = jnp.finfo(preds.dtype).eps
    signal_power = jnp.sum(target_processed**2, axis=-1) + eps
    noise_power = jnp.sum(noise**2, axis=-1) + eps

    snr = 10 * jnp.log10(base.divide_no_nan(signal_power, noise_power))
    return snr

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      targets: jax.Array,
      zero_mean: bool = False,
  ) -> 'SNR':
    """Computes SNR for a batch of audio signals and creates an SNR metric instance.

    Args:
        predictions: A JAX array of predicted audio signals.
        targets: A JAX array of ground truth audio signals.
        zero_mean: If True, subtracts the mean from the signal and noise before
          calculating their respective powers.

    Returns:
        An SNR instance containing the SNR value for the current batch,
        ready for averaging.
    """
    batch_snr_value = cls._calculate_snr(
        predictions,
        targets,
        zero_mean=zero_mean,
    )
    return super().from_model_output(values=batch_snr_value)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of base metrics for metrax."""

from clu import metrics as clu_metrics
import flax
import jax
import jax.numpy as jnp


def divide_no_nan(x: jax.Array, y: jax.Array) -> jax.Array:
  """Computes a safe divide which returns 0 if the y is zero."""
  return jnp.where(y != 0, jnp.divide(x, y), 0.0)


@flax.struct.dataclass
class Average(clu_metrics.Average):
  r"""Average Metric inherits clu.metrics.Average and performs safe division."""

  @classmethod
  def from_model_output(
      cls,
      values: jax.Array,
      sample_weights: jax.Array | None = None,
  ) -> 'Average':
    """Updates the metric.

    Args:
      values: A floating point 1D vector representing the values. The shape
      should be (batch_size,).
      sample_weights: An optional floating point 1D vector representing the
        weight of each sample. The shape should be (batch_size,).

    Returns:
      Updated Average metric.
    """
    total = values
    count = jnp.ones_like(values, dtype=values.dtype)
    if sample_weights is not None:
      total = values * sample_weights
      count = count * sample_weights
    return cls(
        total=total.sum(),
        count=count.sum(),
    )

  def compute(self) -> jax.Array:
    return divide_no_nan(self.total, self.count)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metrax LoggingBackend implemenation for Weight & Bias."""

import datetime
import logging

import jax
import numpy as np

_DEFAULT_STEP = 0


def _get_step(kwargs: dict[str, str | int]) -> int:
  """Returns the step from the kwargs, or 0 if not provided."""
  step = kwargs.get("step")
  return _DEFAULT_STEP if step is None else int(step)


def _preprocess_event_name(event_name: str) -> str:
  """Preprocesses the event name before logging."""
  return event_name.lstrip("/")  # Remove leading slashes


class WandbBackend:
  """A logging backend for W&B that conforms to the LoggingBackend protocol."""

  def __init__(self, project: str, name: str | None = None, **kwargs):
    if jax.process_index() != 0:
      self._is_active = False
      self.wandb = None  # Ensure the attribute exists
      return

    try:
      # pylint: disable=g-import-not-at-top
      # pytype: disable=import-error
      import wandb
    except ImportError as e:
      raise ImportError(
          "The 'wandb' library is not installed. Please install it with "
          "'pip install wandb' to use the WandbBackend."
      ) from e
    self.wandb = wandb

    run_name = name or datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    wandb.init(project=project, name=run_name, anonymous="allow", **kwargs)
    if wandb.run:
      logging.info("W&B run URL: %s", wandb.run.url)
      self._is_active = True
    else:
      self._is_active = False

  def log_scalar(self, event: str, value: float | np.ndarray, **kwargs):
    if self.wandb is None or not self._is_active:
      return
    current_step = _get_step(kwargs)
    event_name = _preprocess_event_name(event)
    self.wandb.log({event_name: value}, step=current_step)

  def close(self):
    if self.wandb is None or not self._is_active:
      return
    if hasattr(self.wandb, "run") and self.wandb.run:
      self.wandb.finish()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import metrax
from metrax.nnx import nnx_wrapper

NnxWrapper = nnx_wrapper.NnxWrapper


class AUCPR(NnxWrapper):
  """An NNX class for the Metrax metric AUCPR."""

  def __init__(self):
    super().__init__(metrax.AUCPR)


class AUCROC(NnxWrapper):
  """An NNX class for the Metrax metric AUCROC."""

  def __init__(self):
    super().__init__(metrax.AUCROC)


class Accuracy(NnxWrapper):
  """An NNX class for the Metrax metric Accuracy."""

  def __init__(self):
    super().__init__(metrax.Accuracy)


class Average(NnxWrapper):
  """An NNX class for the Metrax metric Average."""

  def __init__(self):
    super().__init__(metrax.Average)


class AveragePrecisionAtK(NnxWrapper):
  """An NNX class for the Metrax metric AveragePrecisionAtK."""

  def __init__(self):
    super().__init__(metrax.AveragePrecisionAtK)


class BLEU(NnxWrapper):
  """An NNX class for the Metrax metric BLEU."""

  def __init__(self):
    super().__init__(metrax.BLEU)


class DCGAtK(NnxWrapper):
  """An NNX class for the Metrax metric DCGAtK."""

  def __init__(self):
    super().__init__(metrax.DCGAtK)


class Dice(NnxWrapper):
  """An NNX class for the Metrax metric Dice."""

  def __init__(self):
    super().__init__(metrax.Dice)

class FBetaScore(NnxWrapper):
  """An NNX class for the Metrax metric FBetaScore."""

  def __init__(self):
    super().__init__(metrax.FBetaScore)

class IoU(NnxWrapper):
  """An NNX class for the Metrax metric IoU."""

  def __init__(self):
    super().__init__(metrax.IoU)


class MAE(NnxWrapper):
  """An NNX class for the Metrax metric MAE."""

  def __init__(self):
    super().__init__(metrax.MAE)


class MRR(NnxWrapper):
  """An NNX class for the Metrax metric MRR."""

  def __init__(self):
    super().__init__(metrax.MRR)


class MSE(NnxWrapper):
  """An NNX class for the Metrax metric MSE."""

  def __init__(self):
    super().__init__(metrax.MSE)


class NDCGAtK(NnxWrapper):
  """An NNX class for the Metrax metric NDCGAtK."""

  def __init__(self):
    super().__init__(metrax.NDCGAtK)


class Perplexity(NnxWrapper):
  """An NNX class for the Metrax metric Perplexity."""

  def __init__(self):
    super().__init__(metrax.Perplexity)


class Precision(NnxWrapper):
  """An NNX class for the Metrax metric Precision."""

  def __init__(self):
    super().__init__(metrax.Precision)


class PrecisionAtK(NnxWrapper):
  """An NNX class for the Metrax metric PrecisionAtK."""

  def __init__(self):
    super().__init__(metrax.PrecisionAtK)


class PSNR(NnxWrapper):
  """An NNX class for the Metrax metric PSNR."""

  def __init__(self):
    super().__init__(metrax.PSNR)


class Recall(NnxWrapper):
  """An NNX class for the Metrax metric Recall."""

  def __init__(self):
    super().__init__(metrax.Recall)


class RecallAtK(NnxWrapper):
  """An NNX class for the Metrax metric RecallAtK."""

  def __init__(self):
    super().__init__(metrax.RecallAtK)


class RMSE(NnxWrapper):
  """An NNX class for the Metrax metric RMSE."""

  def __init__(self):
    super().__init__(metrax.RMSE)


class RougeL(NnxWrapper):
  """An NNX class for the Metrax metric RougeL."""

  def __init__(self):
    super().__init__(metrax.RougeL)


class RougeN(NnxWrapper):
  """An NNX class for the Metrax metric RougeN."""

  def __init__(self):
    super().__init__(metrax.RougeN)


class RSQUARED(NnxWrapper):
  """An NNX class for the Metrax metric RSQUARED."""

  def __init__(self):
    super().__init__(metrax.RSQUARED)


class SNR(NnxWrapper):
  """An NNX class for the Metrax metric SNR."""

  def __init__(self):
    super().__init__(metrax.SNR)


class SSIM(NnxWrapper):
  """An NNX class for the Metrax metric SSIM."""

  def __init__(self):
    super().__init__(metrax.SSIM)


class WER(NnxWrapper):
  """An NNX class for the Metrax metric WER."""

  def __init__(self):
    super().__init__(metrax.WER)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of different metrics for regression models."""

from clu import metrics as clu_metrics
import flax
import jax
import jax.numpy as jnp
from metrax import base


@flax.struct.dataclass
class MAE(base.Average):
  r"""Computes the mean absolute error for regression problems given `predictions` and `labels`.

  The mean absolute error without sample weights is defined as:

  .. math::
      MAE = \frac{1}{N} \sum_{i=1}^{N} |y_i - \hat{y}_i|

  When sample weights :math:`w_i` are provided, the weighted mean absolute error
  is:

  .. math::
      MAE = \frac{\sum_{i=1}^{N} w_i|y_i - \hat{y}_i|}{\sum_{i=1}^{N} w_i}

  where:
      - :math:`y_i` are true values
      - :math:`\hat{y}_i` are predictions
      - :math:`w_i` are sample weights
      - :math:`N` is the number of samples
  """

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
  ) -> 'MAE':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector representing the prediction
        generated from the model. The shape should be (batch_size,).
      labels: True value. The shape should be (batch_size,).
      sample_weights: An optional floating point 1D vector representing the
        weight of each sample. The shape should be (batch_size,).

    Returns:
      Updated MAE metric. The shape should be a single scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    absolute_error = jnp.abs(predictions - labels)
    count = jnp.ones_like(labels, dtype=jnp.int32)
    if sample_weights is not None:
      absolute_error = absolute_error * sample_weights
      count = count * sample_weights
    return cls(
        total=absolute_error.sum(),
        count=count.sum(),
    )


@flax.struct.dataclass
class MSE(base.Average):
  r"""Computes the mean squared error for regression problems given `predictions` and `labels`.

  The mean squared error without sample weights is defined as:

  .. math::
      MSE = \frac{1}{N} \sum_{i=1}^{N} (y_i - \hat{y}_i)^2

  When sample weights :math:`w_i` are provided, the weighted mean squared error
  is:

  .. math::
      MSE = \frac{\sum_{i=1}^{N} w_i(y_i - \hat{y}_i)^2}{\sum_{i=1}^{N} w_i}

  where:
      - :math:`y_i` are true values
      - :math:`\hat{y}_i` are predictions
      - :math:`w_i` are sample weights
      - :math:`N` is the number of samples
  """

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
  ) -> 'MSE':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector representing the prediction
        generated from the model. The shape should be (batch_size,).
      labels: True value. The shape should be (batch_size,).
      sample_weights: An optional floating point 1D vector representing the
        weight of each sample. The shape should be (batch_size,).

    Returns:
      Updated MSE metric. The shape should be a single scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    squared_error = jnp.square(predictions - labels)
    count = jnp.ones_like(labels, dtype=jnp.int32)
    if sample_weights is not None:
      squared_error = squared_error * sample_weights
      count = count * sample_weights
    return cls(
        total=squared_error.sum(),
        count=count.sum(),
    )


@flax.struct.dataclass
class RMSE(MSE):
  r"""Computes the root mean squared error for regression problems given `predictions` and `labels`.

  The root mean squared error without sample weights is defined as:

  .. math::
      RMSE = \sqrt{\frac{1}{N} \sum_{i=1}^{N} (y_i - \hat{y}_i)^2}

  When sample weights :math:`w_i` are provided, the weighted root mean squared
  error is:

  .. math::
      RMSE = \sqrt{\frac{\sum_{i=1}^{N} w_i(y_i - \hat{y}_i)^2}{\sum_{i=1}^{N}
      w_i}}

  where:
      - :math:`y_i` are true values
      - :math:`\hat{y}_i` are predictions
      - :math:`w_i` are sample weights
      - :math:`N` is the number of samples
  """

  def compute(self) -> jax.Array:
    return jnp.sqrt(super().compute())


@flax.struct.dataclass
class RSQUARED(clu_metrics.Metric):
  r"""Computes the r-squared score of a scalar or a batch of tensors.

  R-squared is a measure of how well the regression model fits the data. It
  measures the proportion of the variance in the dependent variable that is
  explained by the independent variable(s). It is defined as 1 - SSE / SST,
  where SSE is the sum of squared errors and SST is the total sum of squares.

  .. math::
      R^2 = 1 - \frac{SSE}{SST}

  where:
      .. math::
          SSE = \sum_{i=1}^{N} (y_i - \hat{y}_i)^2
      .. math::
          SST = \sum_{i=1}^{N} (y_i - \bar{y})^2

  When sample weights :math:`w_i` are provided:

  .. math::
      R^2 = 1 - \frac{\sum_{i=1}^{N} w_i(y_i - \hat{y}_i)^2}{\sum_{i=1}^{N}
      w_i(y_i - \bar{y})^2}

  where:
      - :math:`y_i` are true values
      - :math:`\hat{y}_i` are predictions
      - :math:`\bar{y}` is the mean of true values
      - :math:`w_i` are sample weights
      - :math:`N` is the number of samples

  The score ranges from -∞ to 1, where 1 indicates perfect prediction and 0
  indicates
  that the model performs no better than a horizontal line.
  """

  total: jax.Array
  count: jax.Array
  sum_of_squared_error: jax.Array
  sum_of_squared_label: jax.Array

  @classmethod
  def empty(cls) -> 'RSQUARED':
    return cls(
        total=jnp.array(0, jnp.float32),
        count=jnp.array(0, jnp.float32),
        sum_of_squared_error=jnp.array(0, jnp.float32),
        sum_of_squared_label=jnp.array(0, jnp.float32),
    )

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
  ) -> 'RSQUARED':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector representing the prediction
        generated from the model. The shape should be (batch_size,).
      labels: True value. The shape should be (batch_size,).
      sample_weights: An optional floating point 1D vector representing the
        weight of each sample. The shape should be (batch_size,).

    Returns:
      Updated RSQUARED metric. The shape should be a single scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    count = jnp.ones_like(labels, dtype=jnp.int32)
    squared_error = jnp.power(labels - predictions, 2)
    squared_label = jnp.power(labels, 2)
    if sample_weights is not None:
      labels = labels * sample_weights
      count = count * sample_weights
      squared_error = squared_error * sample_weights
      squared_label = squared_label * sample_weights
    return cls(
        total=labels.sum(),
        count=count.sum(),
        sum_of_squared_error=squared_error.sum(),
        sum_of_squared_label=squared_label.sum(),
    )

  def merge(self, other: 'RSQUARED') -> 'RSQUARED':
    return type(self)(
        total=self.total + other.total,
        count=self.count + other.count,
        sum_of_squared_error=self.sum_of_squared_error
        + other.sum_of_squared_error,
        sum_of_squared_label=self.sum_of_squared_label
        + other.sum_of_squared_label,
    )

  def compute(self) -> jax.Array:
    r"""Computes the r-squared score.

    Since we don't know the mean of the labels before we aggregate all of the
    data, we will manipulate the formula to be:
    sst = \sum_i (x_i - mean)^2
        = \sum_i (x_i^2 - 2 x_i mean + mean^2)
        = \sum_i x_i^2 - 2 mean \sum_i x_i + N * mean^2
        = \sum_i x_i^2 - 2 mean * N * mean + N * mean^2
        = \sum_i x_i^2 - N * mean^2

    Returns:
      The r-squared score.
    """
    mean = base.divide_no_nan(self.total, self.count)
    sst = self.sum_of_squared_label - self.count * jnp.power(mean, 2)
    return 1 - base.divide_no_nan(self.sum_of_squared_error, sst)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of different metrics for classification models."""

from clu import metrics as clu_metrics
import flax
import jax
import jax.numpy as jnp
from metrax import base


def _default_threshold(num_thresholds: int) -> jax.Array:
  """Returns evenly distributed `num_thresholds` between 0.0 and 1.0.

  Args:
    num_thresholds: The number of thresholds to return.

  Returns:
    Evently distributed `num_thresholds` between 0.0 and 1.0.
  """
  if num_thresholds < 2:
    raise ValueError(
        f'num_thresholds must be at least 2, but got {num_thresholds}.'
    )
  epsilon = 1e-5
  thresholds = jnp.arange(num_thresholds, dtype=jnp.float32) / (
      num_thresholds - 1
  )
  thresholds = thresholds.at[0].set(-epsilon)
  thresholds = thresholds.at[-1].set(1.0 + epsilon)
  return thresholds


@flax.struct.dataclass
class Accuracy(base.Average):
  r"""Computes accuracy, which is the frequency with which `predictions` match `labels`.

  This metric calculates the proportion of correct predictions by comparing
  `predictions` and `labels` element-wise. It is the ratio of the sum of
  weighted correct predictions to the sum of all corresponding weights.
  If no `sample_weights` are provided, weights default to 1 for each element.

  The calculation is as follows:

  .. math::
      \text{Accuracy} = \frac{\sum (\text{weight} \times \text{correct})}{\sum
      \text{weight}}

  where `correct` is 1 if `prediction == label` for an element, and 0 otherwise.
  `weight` is the `sample_weight` for that element, or 1 if no weights are
  given.
  """

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
  ) -> 'Accuracy':
    """Updates the metric state with new `predictions` and `labels`.

    This method computes element-wise equality between `predictions` and
    `labels`. The result of this comparison (a boolean array, treated as 1 for
    True and 0 for False) is then used to update the metric's `total` and
    `count`.

    Args:
      predictions: JAX array of predicted values. Expected to have a shape
        compatible with `labels` for element-wise comparison (e.g.,
        `(batch_size,)`, `(batch_size, num_classes)`, `(batch_size,
        sequence_length, num_features)`).
      labels: JAX array of true values. Expected to have a shape compatible with
        `predictions` for element-wise comparison.
      sample_weights: Optional JAX array of weights. If provided, it must be
        broadcastable to the shape of `labels` (which should also be compatible
        with `predictions`' shape).

    Returns:
      An updated instance of `Accuracy` metric.

    Raises:
      ValueError: If JAX operations (like broadcasting or arithmetic) fail due
        to incompatible shapes or types among `predictions`, `labels`, and
        `sample_weights`. For instance, if `predictions` and `labels` shapes
        are not identical and not broadcastable to a common shape for
        comparison, or if `sample_weights` cannot be broadcast to `labels`'
        shape.
    """
    correct = predictions == labels
    count = jnp.ones_like(labels, dtype=jnp.int32)
    if sample_weights is not None:
      correct = correct * sample_weights
      count = count * sample_weights
    return cls(
        total=correct.sum(),
        count=count.sum(),
    )


@flax.struct.dataclass
class Precision(clu_metrics.Metric):
  r"""Computes precision for binary classification given `predictions` and `labels`.

  It is calculated as:

  .. math::
      Precision = \frac{TP}{TP + FP}

  where:
      - TP (True Positives): Number of correctly predicted positive cases
      - FP (False Positives): Number of incorrectly predicted positive cases

  A threshold parameter (default 0.5) is used to convert probability predictions
  to binary predictions.

  Attributes:
    true_positives: The count of true positive instances from the given data,
      label, and threshold.
    false_positives: The count of false positive instances from the given data,
      label, and threshold.
  """

  true_positives: jax.Array
  false_positives: jax.Array

  @classmethod
  def empty(cls) -> 'Precision':
    return cls(
        true_positives=jnp.array(0, jnp.float32),
        false_positives=jnp.array(0, jnp.float32),
    )

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      threshold: float = 0.5,
  ) -> 'Precision':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector whose values are in the range [0,
        1]. The shape should be (batch_size,).
      labels: True value. The value is expected to be 0 or 1. The shape should
        be (batch_size,).
      threshold: The threshold to use for the binary classification.

    Returns:
      Updated Precision metric. The shape should be a single scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    predictions = jnp.where(predictions >= threshold, 1, 0)
    true_positives = jnp.sum((predictions == 1) & (labels == 1))
    false_positives = jnp.sum((predictions == 1) & (labels == 0))

    return cls(true_positives=true_positives, false_positives=false_positives)

  def merge(self, other: 'Precision') -> 'Precision':
    return type(self)(
        true_positives=self.true_positives + other.true_positives,
        false_positives=self.false_positives + other.false_positives,
    )

  def compute(self) -> jax.Array:
    return base.divide_no_nan(
        self.true_positives, (self.true_positives + self.false_positives)
    )


@flax.struct.dataclass
class Recall(clu_metrics.Metric):
  r"""Computes recall for binary classification given `predictions` and `labels`.

  It is calculated as:

  .. math::
      Recall = \frac{TP}{TP + FN}

  where:
      - TP (True Positives): Number of correctly predicted positive cases
      - FN (False Negatives): Number of incorrectly predicted negative cases

  A threshold parameter (default 0.5) is used to convert probability predictions
  to binary predictions.

  Attributes:
    true_positives: The count of true positive instances from the given data,
      label, and threshold.
    false_negatives: The count of false negative instances from the given data,
      label, and threshold.
  """

  true_positives: jax.Array
  false_negatives: jax.Array

  @classmethod
  def empty(cls) -> 'Recall':
    return cls(
        true_positives=jnp.array(0, jnp.float32),
        false_negatives=jnp.array(0, jnp.float32),
    )

  @classmethod
  def from_model_output(
      cls, predictions: jax.Array, labels: jax.Array, threshold: float = 0.5
  ) -> 'Recall':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector whose values are in the range [0,
        1]. The shape should be (batch_size,).
      labels: True value. The value is expected to be 0 or 1. The shape should
        be (batch_size,).
      threshold: The threshold to use for the binary classification.

    Returns:
      Updated Recall metric. The shape should be a single scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    predictions = jnp.where(predictions >= threshold, 1, 0)
    true_positives = jnp.sum((predictions == 1) & (labels == 1))
    false_negatives = jnp.sum((predictions == 0) & (labels == 1))

    return cls(true_positives=true_positives, false_negatives=false_negatives)

  def merge(self, other: 'Recall') -> 'Recall':
    return type(self)(
        true_positives=self.true_positives + other.true_positives,
        false_negatives=self.false_negatives + other.false_negatives,
    )

  def compute(self) -> jax.Array:
    return base.divide_no_nan(
        self.true_positives, (self.true_positives + self.false_negatives)
    )


@flax.struct.dataclass
class AUCPR(clu_metrics.Metric):
  r"""Computes area under the precision-recall curve for binary classification given `predictions` and `labels`.

  The Precision-Recall curve shows the tradeoff between precision and recall at
  different
  classification thresholds. The area under this curve (AUC-PR) provides a
  single score
  that represents the model's ability to identify positive cases across
  all possible classification thresholds, particularly in imbalanced datasets.

  For each threshold :math:`t`, precision and recall are calculated as:

  .. math::
    Precision(t) = \frac{TP(t)}{TP(t) + FP(t)}

    Recall(t) = \frac{TP(t)}{TP(t) + FN(t)}

  The AUC-PR is then computed using interpolation:

  .. math::
    AUC-PR = \sum_{i=1}^{n-1} (R_{i+1} - R_i) \cdot \frac{P_i + P_{i+1}}{2}

  where:
    - :math:`P_i` is precision at threshold i
    - :math:`R_i` is recall at threshold i
    - :math:`n` is the number of thresholds

  AUC-PR Curve metric have a number of known issues so use it with caution.
  - PR curves are highly class balance sensitive.
  - PR is a non-monotonic function and thus its "area" is not directly
    proportional to performance.
  - PR-AUC has no standard implementation and different libraries will give
    different results. Some libraries will interpolate between points, others
    will assume a step function (or trapezoidal as sklearn does). Some libraries
    will compute the convex hull of the PR curve, others will not. Because PR is
    non monotonic, its value is sensitive to the number of samples along the
    curve (more so than ROC-AUC).

  Attributes:
    true_positives: The count of true positive instances from the given data and
      label at each threshold.
    false_positives: The count of false positive instances from the given data
      and label at each threshold.
    false_negatives: The count of false negative instances from the given data
      and label at each threshold.
  """

  # shape: (threshold, 1)
  true_positives: jax.Array
  false_positives: jax.Array
  false_negatives: jax.Array
  num_thresholds: int

  @classmethod
  def empty(cls) -> 'AUCPR':
    return cls(
        true_positives=jnp.array(0, jnp.float32),
        false_positives=jnp.array(0, jnp.float32),
        false_negatives=jnp.array(0, jnp.float32),
        num_thresholds=0,
    )

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
      num_thresholds: int = 200,
  ) -> 'AUCPR':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector whose values are in the range [0,
        1]. The shape should be (batch_size,).
      labels: True value. The value is expected to be 0 or 1. The shape should
        be (batch_size,).
      sample_weights: An optional floating point 1D vector representing the
        weight of each sample. The shape should be (batch_size,).
      num_thresholds: The number of thresholds to use. Default is 200.

    Returns:
      The area under the precision-recall curve. The shape should be a single
      scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    pred_is_pos = jnp.greater(
        predictions,
        _default_threshold(num_thresholds=num_thresholds)[..., None],
    )
    pred_is_neg = jnp.logical_not(pred_is_pos)
    label_is_pos = jnp.equal(labels, 1)
    label_is_neg = jnp.equal(labels, 0)

    true_positives = pred_is_pos * label_is_pos
    false_positives = pred_is_pos * label_is_neg
    false_negatives = pred_is_neg * label_is_pos

    if sample_weights is not None:
      true_positives = true_positives * sample_weights
      false_positives = false_positives * sample_weights
      false_negatives = false_negatives * sample_weights

    return cls(
        true_positives=true_positives.sum(axis=-1),
        false_positives=false_positives.sum(axis=-1),
        false_negatives=false_negatives.sum(axis=-1),
        num_thresholds=num_thresholds,
    )

  def merge(self, other: 'AUCPR') -> 'AUCPR':
    return type(self)(
        true_positives=self.true_positives + other.true_positives,
        false_positives=self.false_positives + other.false_positives,
        false_negatives=self.false_negatives + other.false_negatives,
        num_thresholds=self.num_thresholds,
    )

  def interpolate_pr_auc(self) -> jax.Array:
    """Interpolation formula inspired by section 4 of Davis & Goadrich 2006.

    https://minds.wisconsin.edu/handle/1793/60482

    Note here we derive & use a closed formula not present in the paper
    as follows:

      Precision = TP / (TP + FP) = TP / P

    Modeling all of TP (true positive), FP (false positive) and their sum
    P = TP + FP (predicted positive) as varying linearly within each
    interval [A, B] between successive thresholds, we get

      Precision slope = dTP / dP
                      = (TP_B - TP_A) / (P_B - P_A)
                      = (TP - TP_A) / (P - P_A)
      Precision = (TP_A + slope * (P - P_A)) / P

    The area within the interval is (slope / total_pos_weight) times

      int_A^B{Precision.dP} = int_A^B{(TP_A + slope * (P - P_A)) * dP / P}
      int_A^B{Precision.dP} = int_A^B{slope * dP + intercept * dP / P}

    where intercept = TP_A - slope * P_A = TP_B - slope * P_B, resulting in

      int_A^B{Precision.dP} = TP_B - TP_A + intercept * log(P_B / P_A)

    Bringing back the factor (slope / total_pos_weight) we'd put aside, we
    get

      slope * [dTP + intercept *  log(P_B / P_A)] / total_pos_weight

    where dTP == TP_B - TP_A.

    Note that when P_A == 0 the above calculation simplifies into

      int_A^B{Precision.dTP} = int_A^B{slope * dTP} = slope * (TP_B - TP_A)

    which is really equivalent to imputing constant precision throughout the
    first bucket having >0 true positives.

    Returns:
      pr_auc: A float scalar jax.Array that is an approximation of the area
      under the P-R curve.
    """
    dtp = (
        self.true_positives[: self.num_thresholds - 1] - self.true_positives[1:]
    )
    p = self.true_positives + self.false_positives
    dp = p[: self.num_thresholds - 1] - p[1:]
    prec_slope = base.divide_no_nan(dtp, jnp.maximum(dp, 0))
    intercept = self.true_positives[1:] - prec_slope * p[1:]

    # recall_relative_ratio
    safe_p_ratio = jnp.where(
        jnp.multiply(p[: self.num_thresholds - 1] > 0, p[1:] > 0),
        base.divide_no_nan(
            p[: self.num_thresholds - 1],
            jnp.maximum(p[1:], 0),
        ),
        jnp.ones_like(p[1:]),
    )
    # pr_auc_increment
    pr_auc_increment = base.divide_no_nan(
        prec_slope * (dtp + intercept * jnp.log(safe_p_ratio)),
        jnp.maximum(self.true_positives[1:] + self.false_negatives[1:], 0),
    )
    return jnp.sum(pr_auc_increment)

  def compute(self) -> jax.Array:
    # Use interpolation to compute the area under the PR curve to match Keras.
    return self.interpolate_pr_auc()


@flax.struct.dataclass
class AUCROC(clu_metrics.Metric):
  r"""Computes area under the receiver operation characteristic curve for binary classification given `predictions`
  and `labels`.

  The ROC curve shows the tradeoff between the true positive rate (TPR) and
  false positive
  rate (FPR) at different classification thresholds. The area under this curve
  (AUC-ROC)
  provides a single score that represents the model's ability to discriminate
  between
  positive and negative cases across all possible classification thresholds,
  regardless of class imbalance.

  For each threshold :math:`t`, TPR and FPR are calculated as:

  .. math::
      TPR(t) = \frac{TP(t)}{TP(t) + FN(t)}

      FPR(t) = \frac{FP(t)}{FP(t) + TN(t)}

  The AUC-ROC is then computed using the trapezoidal rule:

  .. math::
      AUC-ROC = \int_{0}^{1} TPR(FPR^{-1}(x)) dx

  A score of 1 represents perfect classification, while 0.5 represents random
  guessing.

  Attributes:
    true_positives: The count of true positive instances from the given data and
      label at each threshold.
    false_positives: The count of false positive instances from the given data
      and label at each threshold.
    total_count: The count of every data point.
  """

  # shape: (threshold, 1)
  true_positives: jax.Array
  true_negatives: jax.Array
  false_positives: jax.Array
  false_negatives: jax.Array
  num_thresholds: int

  @classmethod
  def empty(cls) -> 'AUCROC':
    return cls(
        true_positives=jnp.array(0, jnp.float32),
        true_negatives=jnp.array(0, jnp.float32),
        false_positives=jnp.array(0, jnp.float32),
        false_negatives=jnp.array(0, jnp.float32),
        num_thresholds=0,
    )

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
      num_thresholds: int = 200,
  ) -> 'AUCROC':
    """Updates the metric.

    Args:
      predictions: A floating point 1D vector whose values are in the range [0,
        1]. The shape should be (batch_size,).
      labels: True value. The value is expected to be 0 or 1. The shape should
        be (batch_size,).
      sample_weights: An optional floating point 1D vector representing the
        weight of each sample. The shape should be (batch_size,).
      num_thresholds: The number of thresholds to use. Default is 200.

    Returns:
      The area under the receiver operation characteristic curve. The shape
      should be a single scalar.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    pred_is_pos = jnp.greater(
        predictions,
        _default_threshold(num_thresholds=num_thresholds)[..., None],
    )
    pred_is_neg = jnp.logical_not(pred_is_pos)
    label_is_pos = jnp.equal(labels, 1)
    label_is_neg = jnp.equal(labels, 0)

    true_positives = pred_is_pos * label_is_pos
    true_negatives = pred_is_neg * label_is_neg
    false_positives = pred_is_pos * label_is_neg
    false_negatives = pred_is_neg * label_is_pos

    if sample_weights is not None:
      true_positives = true_positives * sample_weights
      true_negatives = true_negatives * sample_weights
      false_positives = false_positives * sample_weights
      false_negatives = false_negatives * sample_weights

    return cls(
        true_positives=true_positives.sum(axis=-1),
        true_negatives=true_negatives.sum(axis=-1),
        false_positives=false_positives.sum(axis=-1),
        false_negatives=false_negatives.sum(axis=-1),
        num_thresholds=num_thresholds,
    )

  def merge(self, other: 'AUCROC') -> 'AUCROC':
    return type(self)(
        true_positives=self.true_positives + other.true_positives,
        true_negatives=self.true_negatives + other.true_negatives,
        false_positives=self.false_positives + other.false_positives,
        false_negatives=self.false_negatives + other.false_negatives,
        num_thresholds=self.num_thresholds,
    )

  def compute(self) -> jax.Array:
    tp_rate = base.divide_no_nan(
        self.true_positives, self.true_positives + self.false_negatives
    )
    fp_rate = base.divide_no_nan(
        self.false_positives, self.false_positives + self.true_negatives
    )
    # Threshold goes from 0 to 1, so trapezoid is negative.
    return jnp.trapezoid(tp_rate, fp_rate) * -1

@flax.struct.dataclass
class FBetaScore(clu_metrics.Metric):
    """
    F-Beta score Metric class
    Computes the F-Beta score for the binary classification given 'predictions' and 'labels'.

    Formula for F-Beta Score:
        b2 = beta ** 2
        f_beta_score = ((1 + b2) * (precision * recall)) / (precision * b2 + recall)

    F-Beta turns into the F1 Score when beta = 1.0

    Attributes:
        true_positives: The count of true positive instances from the given data,
            label, and threshold.
        false_positives: The count of false positive instances from the given data,
            label, and threshold.
        false_negatives: The count of false negative instances from the given data,
            label, and threshold.
        beta: The beta value used in the F-Score metric
    """

    true_positives: jax.Array
    false_positives: jax.Array
    false_negatives: jax.Array
    beta: float = 1.0

    # Reset the variables for the class
    @classmethod
    def empty(cls) -> 'FBetaScore':
        return cls(
            true_positives = jnp.array(0, jnp.float32),
            false_positives = jnp.array(0, jnp.float32),
            false_negatives = jnp.array(0, jnp.float32),
            beta=1.0,
        )

    @classmethod
    def from_model_output(
            cls,
            predictions: jax.Array,
            labels: jax.Array,
            beta = beta,
            threshold = 0.5,) -> 'FBetaScore':
        """Updates the metric.
            Note: When only predictions and labels are given, the score calculated
            is the F1 score if the FBetaScore beta value has not been previously modified.

            Args:
                predictions: A floating point 1D vector whose values are in the range [0,
                  1]. The shape should be (batch_size,).
                labels: True value. The value is expected to be 0 or 1. The shape should
                  be (batch_size,).
                beta: beta value to use in the F-Score metric. A floating number.
                threshold: threshold value to use in the F-Score metric. A floating number.

            Returns:
                The updated FBetaScore object.

            Raises:
                ValueError: If type of `labels` is wrong or the shapes of `predictions`
                and `labels` are incompatible. If the beta or threshold are invalid values
                an error is raised as well.
        """

        # Ensure that beta is a floating number and a valid number
        if not isinstance(beta, float):
            raise ValueError('The "Beta" value must be a floating number.')
        if beta <= 0.0:
            raise ValueError('The "Beta" value must be larger than 0.0.')

        # Ensure that threshold is a floating number and a valid number
        if not isinstance(threshold, float):
            raise ValueError('The "Threshold" value must be a floating number.')
        if threshold < 0.0 or threshold > 1.0:
            raise ValueError('The "Threshold" value must be between 0 and 1.')

        # Modify predictions with the given threshold value
        predictions = jnp.where(predictions >= threshold, 1, 0)

        # Assign the true_positive, false_positive, and false_negative their values
        """
        We are calculating these values manually instead of using Metrax's
        precision and recall classes. This is because the Metrax versions end up
        outputting a single numerical answer when we need an array of numbers.
        """
        true_positives = jnp.sum(predictions * labels, axis = 0)
        false_positives = jnp.sum(predictions * (1 - labels), axis = 0)
        false_negatives = jnp.sum((1- predictions) * labels, axis = 0)

        return cls(true_positives = true_positives,
                   false_positives = false_positives,
                   false_negatives = false_negatives,
                   beta = beta)

    """
    This function is currently unused as the 'from_model_output' function can handle the whole
    dataset without needing to split and merge them. I'm leaving this here for now incase we want to
    repurpose this or need to change something that requires this function's use again. This function would need
    to be reworked for it to work with the current implementation of this class.
    """
    # # Merge datasets together
    # def merge(self, other: 'FBetaScore') -> 'FBetaScore':
    #
    #     # Check if the incoming beta is the same value as the current beta
    #     if other.beta == self.beta:
    #         return type(self)(
    #             true_positives = self.true_positives + other.true_positives,
    #             false_positives = self.false_positives + other.false_positives,
    #             false_negatives = self.false_negatives + other.false_negatives,
    #             beta=self.beta,
    #         )
    #     else:
    #         raise ValueError('The "Beta" values between the two are not equal.')

    # Compute the F-Beta score metric
    def compute(self) -> jax.Array:

        # Epsilon fuz factor required to match with the keras version
        epsilon = 1e-7

        # Manually calculate precision and recall
        """
        This is done in this manner since the metrax variants of precision
        and recall output only single numbers. To match the keras value
        we need an array of numbers to work with.
        """
        precision = jnp.divide(
            self.true_positives,
            self.true_positives + self.false_positives + epsilon,
        )
        recall = jnp.divide(
            self.true_positives,
            self.true_positives + self.false_negatives + epsilon,
        )

        # Compute the numerator and denominator of the F-Score formula
        b2 = self.beta ** 2
        numerator = (1 + b2) * (precision * recall)
        denominator = (b2 * precision) + recall

        return base.divide_no_nan(
            numerator, denominator + epsilon,
        )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metrax LoggingBackend implemenation for Tensorboard."""

import jax
import numpy as np
from tensorboardX import writer

_DEFAULT_STEP = 0


def _get_step(kwargs: dict[str, str | int]) -> int:
  """Returns the step from the kwargs, or 0 if not provided."""
  step = kwargs.get("step")
  return _DEFAULT_STEP if step is None else int(step)


def _preprocess_event_name(event_name: str) -> str:
  """Preprocesses the event name before logging."""
  return event_name.lstrip("/")  # Remove leading slashes


class TensorboardBackend:
  """A logging backend for Tensorboard that conforms to the LoggingBackend protocol."""

  def __init__(self, log_dir: str, flush_every_n_steps: int = 100):
    self._flush_every_n_steps = flush_every_n_steps
    if jax.process_index() == 0:
      self._writer = writer.SummaryWriter(logdir=log_dir)
    else:
      self._writer = None

  def log_scalar(self, event: str, value: float | np.ndarray, **kwargs):
    if self._writer is None:
      return
    current_step = _get_step(kwargs)
    event_name = _preprocess_event_name(event)
    self._writer.add_scalar(event_name, value, current_step)
    if current_step % self._flush_every_n_steps == 0:
      self._writer.flush()

  def close(self):
    if self._writer:
      self._writer.close()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A wrapper for metrax metrics to be used with NNX."""

from flax import nnx


class NnxWrapper(nnx.metrics.Metric):
  """A wrapper class for clu metrics to be used with NNX."""

  def __init__(self, cls):
    self.clu_metric = cls.empty()

  def reset(self) -> None:
    self.clu_metric = self.clu_metric.empty()

  def update(self, **kwargs) -> None:
    other_clu_metric = self.clu_metric.from_model_output(**kwargs)
    self.clu_metric = self.clu_metric.merge(other_clu_metric)

  def compute(self):
    return self.clu_metric.compute()

  def __init_subclass__(cls, **kwargs):
    super().__init_subclass__(pytree=False, **kwargs)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of different metrics for image models."""

from clu import metrics as clu_metrics
import flax
import jax
from jax import lax
import jax.numpy as jnp
from metrax import base


def _gaussian_kernel1d(sigma, radius):
  r"""Generates a 1D normalized Gaussian kernel.

  This function creates a 1D Gaussian kernel, which can be used for smoothing
  operations. The kernel is centered at zero and its values are determined by
  the Gaussian function:

  .. math::
    \phi(x) = e^{-\frac{x^2}{2\sigma^2}}

  The resulting kernel :math:`\phi(x)` is then normalized by dividing each
  element by the sum of all elements, so that the sum of the kernel's elements
  is 1. This function assumes an order of 0 for the Gaussian derivative (i.e.,
  a standard smoothing kernel).

  Args:
    sigma (float): The standard deviation (:math:`\sigma`) of the Gaussian
      distribution. This controls the "width" or "spread" of the kernel.
    radius (int): The radius of the kernel. The kernel will include points from
      :math:`-radius` to :math:`+radius`. The total size of the kernel will be
      :math:`2 \times radius + 1`.

  Returns:
    jnp.ndarray: A 1D JAX array representing the normalized Gaussian kernel.
  """
  sigma2 = sigma * sigma
  x = jnp.arange(-radius, radius + 1)
  phi_x = jnp.exp(-0.5 / sigma2 * x**2)
  phi_x = phi_x / phi_x.sum()
  return phi_x


@flax.struct.dataclass
class SSIM(base.Average):
  r"""SSIM (Structural Similarity Index Measure) Metric.

  This class calculates the structural similarity between predicted and target
  images and averages it over a dataset. SSIM is a perception-based model that
  considers changes in structural information, luminance, and contrast.

  The general SSIM formula considers three components: luminance (l),
      contrast (c), and structure (s):

      .. math::
        SSIM(x, y) = [l(x, y)]^\alpha \cdot [c(x, y)]^\beta \cdot [s(x,
        y)]^\gamma

      Where:
        - Luminance comparison:
          :math:`l(x, y) = \frac{2\mu_x\mu_y + c_1}{\mu_x^2 + \mu_y^2 + c_1}`
        - Contrast comparison:
          :math:`c(x, y) = \frac{2\sigma_x\sigma_y + c_2}{\sigma_x^2 +
          \sigma_y^2 + c_2}`
        - Structure comparison:
          :math:`s(x, y) = \frac{\sigma_{xy} + c_3}{\sigma_x\sigma_y + c_3}`

      This implementation uses a common simplified form where :math:`\alpha =
      \beta = \gamma = 1` and :math:`c_3 = c_2 / 2`.

      This leads to the combined formula:

      .. math::
        SSIM(x, y) = \frac{(2\mu_x\mu_y + c_1)(2\sigma_{xy} + c_2)}{(\mu_x^2 +
        \mu_y^2 + c_1)(\sigma_x^2 + \sigma_y^2 + c_2)}

      In these formulas:
        - :math:`\mu_x` and :math:`\mu_y` are the local means of :math:`x` and
        :math:`y`.
        - :math:`\sigma_x^2` and :math:`\sigma_y^2` are the local variances of
        :math:`x` and :math:`y`.
        - :math:`\sigma_{xy}` is the local covariance of :math:`x` and
        :math:`y`.
        - :math:`c_1 = (K_1 L)^2` and :math:`c_2 = (K_2 L)^2` are stabilization
        constants,
          where :math:`L` is the dynamic range of pixel values, and :math:`K_1,
          K_2` are small constants (e.g., 0.01 and 0.03).
  """

  @staticmethod
  def _calculate_ssim(
      img1: jnp.ndarray,
      img2: jnp.ndarray,
      max_val: float,
      filter_size: int = 11,
      filter_sigma: float = 1.5,
      k1: float = 0.01,
      k2: float = 0.03,
  ) -> jnp.ndarray:
    """Computes SSIM (Structural Similarity Index Measure) values for a batch of images.

    This function calculates the SSIM between two batches of images (`img1` and
    `img2`). If the images have multiple channels, SSIM is calculated for each
    channel independently, and then the mean SSIM across channels is returned.

    Args:
      img1: The first batch of images, expected shape ``(batch, height, width,
        channels)``.
      img2: The second batch of images, expected shape ``(batch, height, width,
        channels)``.
      max_val: The dynamic range of the pixel values (e.g., 1.0 for images
        normalized to [0,1] or 255 for uint8 images).
      filter_size: The size of the Gaussian filter window used for calculating
        local statistics. Must be an odd integer.
      filter_sigma: The standard deviation of the Gaussian filter.
      k1: A small constant used in the SSIM formula to stabilize the luminance
        comparison.
      k2: A small constant used in the SSIM formula to stabilize the
        contrast/structure comparison.

    Returns:
      A 1D JAX array of shape ``(batch,)`` containing the SSIM value for each
      image pair in the batch.
    """
    if img1.shape != img2.shape:
      raise ValueError(
          f'Input images must have the same shape, but got {img1.shape} and'
          f' {img2.shape}'
      )
    if img1.ndim != 4:  # (batch, H, W, C)
      raise ValueError(
          'Input images must be 4D tensors (batch, height, width, channels),'
          f' but got {img1.ndim}D'
      )
    if img1.shape[-3] < filter_size or img1.shape[-2] < filter_size:
      raise ValueError(
          f'Image dimensions ({img1.shape[-3]}x{img1.shape[-2]}) must be at'
          f' least filter_size x filter_size ({filter_size}x{filter_size}).'
      )

    num_channels = img1.shape[-1]
    img1 = img1.astype(jnp.float32)
    img2 = img2.astype(jnp.float32)

    gaussian_kernal_1d = _gaussian_kernel1d(
        filter_sigma, (filter_size - 1) // 2
    )
    gaussian_kernel_2d = jnp.outer(gaussian_kernal_1d, gaussian_kernal_1d)
    # Kernel for convolution: (H_k, W_k, C_in=1, C_out=1)
    kernel_conv = gaussian_kernel_2d[:, :, jnp.newaxis, jnp.newaxis]

    c1 = (k1 * max_val) ** 2
    c2 = (k2 * max_val) ** 2

    def _calculate_ssim_for_channel(x_ch, y_ch, conv_kernel, c1, c2):
      r"""Calculates the Structural Similarity Index (SSIM) for a single channel.

      This function computes the SSIM between two single-channel image arrays
      (:math:`x_{ch}` and :math:`y_{ch}`) using a precomputed Gaussian kernel
      for local statistics. The SSIM metric quantifies image quality
      degradation based on perceived changes in structural information, also
      incorporating important perceptual phenomena like luminance and contrast
      masking.

      The general SSIM formula considers three components: luminance (l),
      contrast (c), and structure (s):

      .. math::
        SSIM(x, y) = [l(x, y)]^\alpha \cdot [c(x, y)]^\beta \cdot [s(x,
        y)]^\gamma

      Where:
        - Luminance comparison:
          :math:`l(x, y) = \frac{2\mu_x\mu_y + c_1}{\mu_x^2 + \mu_y^2 + c_1}`
        - Contrast comparison:
          :math:`c(x, y) = \frac{2\sigma_x\sigma_y + c_2}{\sigma_x^2 +
          \sigma_y^2 + c_2}`
        - Structure comparison:
          :math:`s(x, y) = \frac{\sigma_{xy} + c_3}{\sigma_x\sigma_y + c_3}`

      This implementation uses a common simplified form where :math:`\alpha =
      \beta = \gamma = 1` and :math:`c_3 = c_2 / 2`.

      This leads to the combined formula:

      .. math::
        SSIM(x, y) = \frac{(2\mu_x\mu_y + c_1)(2\sigma_{xy} + c_2)}{(\mu_x^2 +
        \mu_y^2 + c_1)(\sigma_x^2 + \sigma_y^2 + c_2)}

      In these formulas:
        - :math:`\mu_x` and :math:`\mu_y` are the local means of :math:`x` and
        :math:`y`.
        - :math:`\sigma_x^2` and :math:`\sigma_y^2` are the local variances of
        :math:`x` and :math:`y`.
        - :math:`\sigma_{xy}` is the local covariance of :math:`x` and
        :math:`y`.
        - :math:`c_1 = (K_1 L)^2` and :math:`c_2 = (K_2 L)^2` are stabilization
        constants,
          where :math:`L` is the dynamic range of pixel values, and :math:`K_1,
          K_2` are small constants (e.g., 0.01 and 0.03).

      Args:
        x_ch (jnp.ndarray): The first input image channel. Expected shape is
          ``(batch, Height, Width, 1)``.
        y_ch (jnp.ndarray): The second input image channel. Expected shape is
          ``(batch, Height, Width, 1)``.
        conv_kernel (jnp.ndarray): The 2D Gaussian kernel, reshaped to 4D, used
          for calculating local windowed statistics (mean, variance,
          covariance). Expected shape is ``(Kernel_H, Kernel_W, 1, 1)``.
        c1 (float): Stabilization constant for the luminance and mean component,
          :math:`(K_1 L)^2`.
        c2 (float): Stabilization constant for the variance and covariance
          component, :math:`(K_2 L)^2`.

      Returns:
        jnp.ndarray: A scalar JAX array (or an array of scalars if batch size >
        1)
        representing the mean SSIM value(s) for the input channel(s).
      """
      # x_ch, y_ch are (batch, H, W, 1)
      dn = lax.conv_dimension_numbers(
          x_ch.shape, conv_kernel.shape, ('NHWC', 'HWIO', 'NHWC')
      )

      mu_x = lax.conv_general_dilated(
          x_ch,
          conv_kernel,
          window_strides=(1, 1),
          padding='VALID',
          dimension_numbers=dn,
      )
      mu_y = lax.conv_general_dilated(
          y_ch,
          conv_kernel,
          window_strides=(1, 1),
          padding='VALID',
          dimension_numbers=dn,
      )

      mu_x_sq = mu_x**2
      mu_y_sq = mu_y**2
      mu_x_mu_y = mu_x * mu_y

      sigma_x_sq = (
          lax.conv_general_dilated(
              x_ch**2,
              conv_kernel,
              window_strides=(1, 1),
              padding='VALID',
              dimension_numbers=dn,
          )
          - mu_x_sq
      )
      sigma_y_sq = (
          lax.conv_general_dilated(
              y_ch**2,
              conv_kernel,
              window_strides=(1, 1),
              padding='VALID',
              dimension_numbers=dn,
          )
          - mu_y_sq
      )
      sigma_xy = (
          lax.conv_general_dilated(
              x_ch * y_ch,
              conv_kernel,
              window_strides=(1, 1),
              padding='VALID',
              dimension_numbers=dn,
          )
          - mu_x_mu_y
      )

      numerator1 = 2 * mu_x_mu_y + c1
      numerator2 = 2 * sigma_xy + c2
      denominator1 = mu_x_sq + mu_y_sq + c1
      denominator2 = sigma_x_sq + sigma_y_sq + c2

      ssim_map = (numerator1 * numerator2) / (denominator1 * denominator2)
      return jnp.mean(
          ssim_map, axis=(1, 2, 3)
      )  # Mean over H, W, C (which is 1 here for the map)

    ssim_per_channel_list = []
    for i in range(num_channels):
      img1_c = lax.dynamic_slice_in_dim(
          img1, i * 1, 1, axis=3
      )  # (batch, H, W, 1)
      img2_c = lax.dynamic_slice_in_dim(
          img2, i * 1, 1, axis=3
      )  # (batch, H, W, 1)

      ssim_for_channel = _calculate_ssim_for_channel(
          img1_c, img2_c, kernel_conv, c1, c2
      )
      ssim_per_channel_list.append(ssim_for_channel)

    ssim_scores_stacked = jnp.stack(
        ssim_per_channel_list, axis=-1
    )  # (batch, num_channels)
    return jnp.mean(ssim_scores_stacked, axis=-1)  # (batch,)

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      targets: jax.Array,
      max_val: float,
      filter_size: int = 11,
      filter_sigma: float = 1.5,
      k1: float = 0.01,
      k2: float = 0.03,
  ) -> 'SSIM':
    """Computes SSIM for a batch of images and creates an SSIM metric instance.

    This method takes batches of predicted and target images, calculates their
    SSIM values, and then initializes an SSIM metric object suitable for
    aggregation across multiple batches.

    Args:
        predictions: A JAX array of predicted images, with shape ``(batch,
          height, width, channels)``.
        targets: A JAX array of ground truth images, with shape ``(batch,
          height, width, channels)``.
        max_val: The maximum possible pixel value (dynamic range) of the images
          (e.g., 1.0 for float images in [0,1], 255 for uint8 images).
        filter_size: The size of the Gaussian filter window used in SSIM
          calculation (default is 11).
        filter_sigma: The standard deviation of the Gaussian filter (default is
          1.5).
        k1: SSIM stability constant for the luminance term (default is 0.01).
        k2: SSIM stability constant for the contrast/structure term (default is
          0.03).

    Returns:
        An SSIM instance containing the SSIM values for the current batch,
        ready for averaging.
    """
    # shape (batch_size,)
    batch_ssim_values = cls._calculate_ssim(
        predictions,
        targets,
        max_val=max_val,
        filter_size=filter_size,
        filter_sigma=filter_sigma,
        k1=k1,
        k2=k2,
    )
    return super().from_model_output(values=batch_ssim_values)


@flax.struct.dataclass
class IoU(base.Average):
  r"""Measures Intersection over Union (IoU) for semantic segmentation.

  The general formula for IoU for a single class is:
  $IoU_{class} = \frac{TP}{TP + FP + FN}$
  where TP, FP, FN are True Positives, False Positives, and False Negatives.

  **Per-Batch Processing:**
  For each input batch, a mean IoU is calculated. This involves:
  1. Aggregating TP, FP, and FN pixel counts for each specified target class
     (from the required `target_class_ids` list) across all samples within the
     batch.
  2. Computing IoU for each of these classes using the batch-aggregated counts:
     $IoU_{class} = \frac{TP}{TP + FP + FN + \epsilon}$.
  3. Averaging these per-class IoU scores to get a single value for the batch.
     - If `target_class_ids` is empty, an array of zeros of shape `(B,)`
       (where `B` is batch size) is produced by `_calculate_iou`.
     - Otherwise, a scalar `jnp.ndarray` (shape `()`) representing the mean
       IoU is produced.

  **Accumulation & Final Metric:**
  This class inherits from `base.Average`. It accumulates the results from
  per-batch processing and `compute()` returns the final mean IoU as a scalar
  `jnp.ndarray` (shape `()`).
  """

  @staticmethod
  def _calculate_iou(
      targets: jnp.ndarray,
      predictions: jnp.ndarray,
      target_class_ids: jnp.ndarray,
      epsilon: float = 1e-7,
  ) -> jnp.ndarray:
    r"""Computes mean IoU for a processed batch by class-wise aggregation using jax.vmap.

    Per-batch processing: For each target class in the provided
    `target_class_ids` list, True Positives (TP), False Positives (FP), and
    False Negatives (FN) are summed across all items in the input batch.
    The IoU for that class is $TP / (TP + FP + FN + \epsilon)$.
    These per-class IoU scores are then averaged. If `target_class_ids` is
    empty, a scalar 0.0 is returned.

    Args:
      targets: Ground truth segmentation masks. Shape is `(B, H, W)`, integer
        class labels. (B: batch size, H: height, W: width)
      predictions: Predicted segmentation masks. Shape is `(B, H, W)`, integer
        class labels.
      target_class_ids: An array of integer class IDs for which to compute IoU.
      epsilon: Small float added to the denominator for numerical stability.
        Default is `1e-7`.

    Returns:
      scalar `jnp.ndarray` (shape `()`) mean IoU for the batch. Returns 0.0
      if `target_class_ids` is empty.
    """
    if target_class_ids.shape[0] == 0:
      return jnp.array(0.0, dtype=jnp.float32)

    def _calculate_iou_for_single_class(
        class_id: jnp.ndarray,
    ) -> jnp.ndarray:
      target_is_class = (targets == class_id)
      pred_is_class = (predictions == class_id)
      intersection = jnp.sum(jnp.logical_and(target_is_class, pred_is_class))
      union = jnp.sum(jnp.logical_or(target_is_class, pred_is_class))
      return intersection / (union + epsilon)

    iou_scores_per_class = jax.vmap(_calculate_iou_for_single_class)(
        target_class_ids
    )

    return jnp.mean(iou_scores_per_class)

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      targets: jax.Array,
      num_classes: int,
      target_class_ids: jax.Array,
      from_logits: bool = False,
      epsilon: float = 1e-7,
  ) -> 'IoU':
    """Creates an `IoU` instance from a batch of model outputs.

    Per-batch processing:
    1. Preprocesses `predictions` and `targets` into integer label masks of
       shape `(B, H, W)`. (B: batch size, H: height, W: width).
    2. Calls `_calculate_iou` using the provided `target_class_ids` to compute
       the batch's mean IoU.

    Args:
      predictions: `jax.Array`. Model predictions. - If `from_logits` is `True`:
        shape `(B, H, W, C)` (C: `num_classes`). - If `from_logits` is `False`:
        shape `(B, H, W)` or `(B, H, W, 1)`.
      targets: `jax.Array`. Ground truth segmentation masks. Shape `(B, H, W)`
        or `(B, H, W, 1)`, integer class labels.
      num_classes: Total number of distinct classes (`C`). Integer.
      target_class_ids: An array of integer class IDs for which to compute IoU.
      from_logits: `bool`. If `True`, `predictions` are logits and argmax is
        applied. Default is `False`.
      epsilon: `float`. Small value for stable IoU calculation. Default is
        `1e-7`.

    Returns:
      An `IoU` metric instance updated with the IoU score from this batch.
    """
    # Preprocessing predictions and targets to be (batch, H, W) integer labels
    if from_logits:
      if predictions.ndim != 4 or predictions.shape[-1] != num_classes:
        raise ValueError(
            'Logit predictions must be 4D (batch, H, W, num_classes) with last'
            f' dim matching num_classes. Got shape {predictions.shape} and'
            f' num_classes {num_classes}'
        )
      processed_predictions = jnp.argmax(predictions, axis=-1).astype(jnp.int32)
    else:
      if predictions.ndim == 4 and predictions.shape[-1] == 1:
        processed_predictions = jnp.squeeze(predictions, axis=-1).astype(
            jnp.int32
        )
      elif predictions.ndim == 3:
        processed_predictions = predictions.astype(jnp.int32)
      else:
        raise ValueError(
            'Predictions (if not from_logits) must be 3D (batch, H, W) or '
            f'4D (batch, H, W, 1). Got shape {predictions.shape}'
        )
    if targets.ndim == 4 and targets.shape[-1] == 1:
      processed_targets = jnp.squeeze(targets, axis=-1).astype(jnp.int32)
    elif targets.ndim == 3:
      processed_targets = targets.astype(jnp.int32)
    else:
      raise ValueError(
          'Targets must be 3D (batch, H, W) or 4D (batch, H, W, 1). '
          f'Got shape {targets.shape}'
      )

    iou_score = cls._calculate_iou(
        targets=processed_targets,
        predictions=processed_predictions,
        target_class_ids=target_class_ids,
        epsilon=epsilon,
    )
    return super().from_model_output(values=iou_score)


@flax.struct.dataclass
class PSNR(base.Average):
  r"""PSNR (Peak Signal-to-Noise Ratio)  Metric.

  This class calculates the Peak Signal-to-Noise Ratio (PSNR) between two images
  to measure the quality of a reconstructed image compared to a reference.

  .. math::

      \text{PSNR}(I, J) = 10 \cdot \log_{10} \left(
      \frac{\max(I)^2}{\text{MSE}(I, J)} \right)

  Where:
    - :math:`\max(I)` is the maximum possible pixel value of the input image.
    - :math:`\text{MSE}(I, J)` is the mean squared error between images
    :math:`I` and :math:`J`.
  """

  @staticmethod
  def _calculate_psnr(
      img1: jnp.ndarray,
      img2: jnp.ndarray,
      max_val: float,
      eps: float = 0,
  ) -> jnp.ndarray:
    """Computes PSNR (Peak Signal-to-Noise Ratio) values.

    Args:
            img1: Predicted images, shape ``(batch, H, W, C)``.
            img2: Ground‑truth images, same shape as ``img1``.
            max_val: Dynamic range of the images (e.g. ``1.0`` or ``255``).
            eps: Small constant to avoid ``log(0)`` when images are identical.

        Returns:
          A 1D JAX array of shape ``(batch,)`` containing PSNR in dB.
    """
    if img1.shape != img2.shape:
      raise ValueError(
          f'Input images must have the same shape, got {img1.shape} and'
          f' {img2.shape}.'
      )
    if img1.ndim != 4:  # (batch, H, W, C)
      raise ValueError(
          'Inputs must be 4‑D (batch, height, width, channels), got'
          f' {img1.ndim}‑D.'
      )

    img1 = img1.astype(jnp.float32)
    img2 = img2.astype(jnp.float32)

    # Mean‑squared error per image.
    mse = jnp.mean(jnp.square(img1 - img2), axis=(1, 2, 3))
    mse = jnp.maximum(mse, eps)

    psnr = 20.0 * jnp.log10(max_val) - 10.0 * jnp.log10(mse)
    return psnr

  @classmethod
  def from_model_output(
      cls,
      predictions: jnp.ndarray,
      targets: jnp.ndarray,
      max_val: float,
  ) -> 'PSNR':
    """Computes PSNR for a batch of images and creates an PSNR metric instance.

    Args:
        predictions: A JAX array of predicted images, with shape ``(batch, H, W,
          C)``.
        targets: A JAX array of ground truth images, with shape ``(batch, H, W,
          C)``.
        max_val: The maximum possible pixel value (dynamic range) of the images
          (e.g., 1.0 for float images in [0,1], 255 for uint8 images).

    Returns:
        A ``PSNR`` instance containing per‑image PSNR values.
    """
    batch_psnr = cls._calculate_psnr(predictions, targets, max_val=max_val)
    return super().from_model_output(values=batch_psnr)


@flax.struct.dataclass
class Dice(clu_metrics.Metric):
  r"""Computes the Dice coefficient between `y_true` and `y_pred`.

  Dice is a similarity measure used to measure overlap between two samples.
  A Dice score of 1 indicates perfect overlap; 0 indicates no overlap.

  The formula is:

  .. math::

      \text{Dice} = \frac{2 \cdot \sum (y_{\text{true}} \cdot y_{\text{pred}})}
                          {\sum y_{\text{true}} + \sum y_{\text{pred}} +
                          \epsilon}

  Attributes:
      intersection: Sum of element-wise product between `y_true` and `y_pred`.
      sum_true: Sum of y_true across all examples.
      sum_pred: Sum of y_pred across all examples.
  """

  intersection: jax.Array
  sum_pred: jax.Array
  sum_true: jax.Array

  @classmethod
  def empty(cls) -> 'Dice':
    return cls(
        intersection=jnp.array(0.0, jnp.float32),
        sum_pred=jnp.array(0.0, jnp.float32),
        sum_true=jnp.array(0.0, jnp.float32),
    )

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
  ) -> 'Dice':
    """Updates the metric.

    Args:
        predictions: A floating point vector whose values are in the range [0,
          1]. The shape should be (batch_size,).
        labels: True value. The value is expected to be 0 or 1. The shape should
          be (batch_size,).

    Returns:
        Updated Dice metric.
    """
    predictions = jnp.asarray(predictions, jnp.float32)
    labels = jnp.asarray(labels, jnp.float32)

    intersection = jnp.sum(predictions * labels)
    sum_pred = jnp.sum(predictions)
    sum_true = jnp.sum(labels)

    return cls(
        intersection=intersection,
        sum_pred=sum_pred,
        sum_true=sum_true,
    )

  def merge(self, other: 'Dice') -> 'Dice':
    return type(self)(
        intersection=self.intersection + other.intersection,
        sum_pred=self.sum_pred + other.sum_pred,
        sum_true=self.sum_true + other.sum_true,
    )

  def compute(self) -> jax.Array:
    """Returns the final Dice coefficient."""
    epsilon = 1e-7
    return (2.0 * self.intersection) / (self.sum_pred + self.sum_true + epsilon)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of different metrics for NLP models."""

import abc
import collections
import math
from clu import metrics as clu_metrics
import flax
import jax
import jax.numpy as jnp
from metrax import base


def _get_single_n_grams(segment: list[str], order: int):
  """Generates a counter of n-grams from a list of tokens for a specific n.

  Args:
    segment: list. Text segment from which n-grams will be extracted.
    order: The order of n-grams.

  Returns:
    A collections.Counter mapping n-gram tuples to their counts.
  """
  return collections.Counter(zip(*[segment[i:] for i in range(order)]))


def _get_ngrams(segment: list[str], max_order: int):
  """Extracts all n-grams up to a given maximum order from an input segment.

  Args:
      segment: list. Text segment from which n-grams will be extracted.
      max_order: int. Maximum length in tokens of the n-grams returned by this
        method.

  Returns:
    A collections.Counter mapping n-gram tuples to their counts for all orders.
  """
  ngram_counts = collections.Counter()
  for order in range(1, max_order + 1):
    ngram_counts.update(_get_single_n_grams(segment, order))
  return ngram_counts


def _lcs_length(str1: list[str], str2: list[str]) -> int:
  """Computes the length of the Longest Common Subsequence (LCS)."""
  lengths = [[0 for j in range(len(str2) + 1)] for i in range(len(str1) + 1)]
  for i, x in enumerate(str1):
    for j, y in enumerate(str2):
      if x == y:
        lengths[i + 1][j + 1] = lengths[i][j] + 1
      else:
        lengths[i + 1][j + 1] = max(lengths[i + 1][j], lengths[i][j + 1])
  return lengths[len(str1)][len(str2)]


@flax.struct.dataclass
class BLEU(clu_metrics.Metric):
  r"""Computes the BLEU score for sequence generation.

  BLEU measures the similarity between a machine-generated candidate translation
  and one or more human reference translations, focusing on matching n-grams.

  It's calculated as:

  .. math::
      BLEU = \text{BP} \cdot \exp\left( \sum_{n=1}^{N} w_n \log p_n \right)

  where:
    - :math:`p_n` is the modified n-gram precision for n-grams of order n.
    - :math:`N` is the maximum n-gram order considered (typically 4).
    - :math:`w_n` are weights for each order (typically uniform, 1/N).
    - :math:`\text{BP}` is the Brevity Penalty.

  This implementation uses uniform weights and calculates statistics
  incrementally.

  Attributes:
    max_order: Maximum n-gram order to consider.
    matches_by_order: Accumulated sum of clipped n-gram matches for each order.
    possible_matches_by_order: Accumulated sum of total n-grams in predictions
      for each order.
    translation_length: Accumulated total length of predictions.
    reference_length: Accumulated total 'effective' reference length (closest
      length match for each prediction).
  """

  max_order: int
  matches_by_order: jax.Array
  possible_matches_by_order: jax.Array
  translation_length: jax.Array
  reference_length: jax.Array

  @classmethod
  def empty(cls) -> 'BLEU':
    return cls(
        max_order=4,
        matches_by_order=jnp.array(0, jnp.float32),
        possible_matches_by_order=jnp.array(0, jnp.float32),
        translation_length=jnp.array(0, jnp.float32),
        reference_length=jnp.array(0, jnp.float32),
    )

  @classmethod
  def from_model_output(
      cls,
      predictions: list[str],
      references: list[list[str]],
      max_order: int = 4,
  ) -> 'BLEU':
    """Computes BLEU statistics for a batch of predictions and references.

    Args:
      predictions: A list of predicted strings. The shape should be (batch_size,
        ).
      references: A list of lists of reference strings. The shape should be
        (batch_size, num_references).
      max_order: The maximum order of n-grams to consider.

    Returns:
      A BLEU metric instance containing the statistics for this batch.

    Raises:
      ValueError: If the shapes of `predictions` and `references` are
      incompatible.
    """
    matches_by_order = [0] * max_order
    possible_matches_by_order = [0] * max_order
    pred_length = 0
    ref_length = 0

    for pred, ref_list in zip(predictions, references):
      pred = pred.split()
      ref_list = [r.split() for r in ref_list]
      pred_length += len(pred)
      ref_length += min(len(r) for r in ref_list)
      prediction_ngram_counts = _get_ngrams(pred, max_order)
      reference_ngram_counts = collections.Counter()
      for ref in ref_list:
        reference_ngram_counts |= _get_ngrams(ref, max_order)
      overlap = prediction_ngram_counts & reference_ngram_counts
      for ngram in overlap:
        matches_by_order[len(ngram) - 1] += overlap[ngram]
      for order in range(1, max_order + 1):
        possible_matches = len(pred) - order + 1
        if possible_matches > 0:
          possible_matches_by_order[order - 1] += possible_matches

    return cls(
        max_order=max_order,
        matches_by_order=jnp.array(matches_by_order, dtype=jnp.float32),
        possible_matches_by_order=jnp.array(
            possible_matches_by_order, dtype=jnp.float32
        ),
        translation_length=jnp.array(pred_length, dtype=jnp.float32),
        reference_length=jnp.array(ref_length, dtype=jnp.float32),
    )

  def merge(self, other: 'BLEU') -> 'BLEU':
    if self.max_order != other.max_order:
      raise ValueError(
          'BLEU metrics with different max_order cannot be merged.'
      )
    return type(self)(
        max_order=self.max_order,
        matches_by_order=(self.matches_by_order + other.matches_by_order),
        possible_matches_by_order=(
            self.possible_matches_by_order + other.possible_matches_by_order
        ),
        translation_length=(self.translation_length + other.translation_length),
        reference_length=(self.reference_length + other.reference_length),
    )

  def compute(self) -> jax.Array:
    precisions = [0] * self.max_order
    for i in range(0, self.max_order):
      precisions[i] = base.divide_no_nan(
          self.matches_by_order[i], self.possible_matches_by_order[i]
      )
    geo_mean = (
        math.exp(sum((1.0 / self.max_order) * math.log(p) for p in precisions))
        if precisions and min(precisions) > 0
        else 0
    )
    ratio = base.divide_no_nan(self.translation_length, self.reference_length)
    bp = 1.0 if ratio > 1.0 else math.exp(1 - 1.0 / ratio)
    bleu = geo_mean * bp
    return jnp.array(bleu)


@flax.struct.dataclass
class Perplexity(clu_metrics.Metric):
  r"""Computes perplexity for sequence generation.

  Perplexity is a measurement of how well a probability distribution predicts a
  sample. It is defined as the exponentiation of the cross-entropy. A low
  perplexity indicates the probability distribution is good at predicting the
  sample.

  For language models, it can be interpreted as the weighted average branching
  factor of the model - how many equally likely words can be selected at each
  step.

  Given a sequence of :math:`N` tokens, perplexity is calculated as:

  .. math::
      Perplexity = \exp\left(-\frac{1}{N}\sum_{i=1}^{N} \log P(x_i|x_{<i})\right)

  When sample weights :math:`w_i` are provided:

  .. math::
      Perplexity = \exp\left(-\frac{\sum_{i=1}^{N} w_i\log P(x_i|x_{<i})}{\sum_{i=1}^{N} w_i}\right)

  where:
      - :math:`P(x_i|x_{<i})` is the predicted probability of token :math:`x_i`
        given previous tokens
      - :math:`w_i` are sample weights
      - :math:`N` is the sequence length

  Lower perplexity indicates better prediction - the model is less "perplexed" by the data.
  """

  aggregate_crossentropy: jax.Array
  num_samples: jax.Array

  @classmethod
  def empty(cls) -> 'Perplexity':
    return cls(
      aggregate_crossentropy=jnp.array(0, jnp.float32),
      num_samples=jnp.array(0, jnp.float32))

  @classmethod
  def from_model_output(
      cls,
      predictions: jax.Array,
      labels: jax.Array,
      sample_weights: jax.Array | None = None,
      from_logits: bool = False,
  ) -> 'Perplexity':
    """Updates the metric.

    Args:
      predictions: A floating point tensor representing the prediction
      generated from the model. The shape should be (batch_size, seq_len,
      vocab_size).
      labels: True value. The shape should be (batch_size, seq_len).
      sample_weights: An optional tensor representing the
        weight of each token. The shape should be (batch_size, seq_len).
      from_logits: Whether the predictions are logits. If True, the predictions
        are converted to probabilities using a softmax. If False, all values
        outside of [0, 1] are clipped to 0 or 1.

    Returns:
      Updated Perplexity metric.

    Raises:
      ValueError: If type of `labels` is wrong or the shapes of `predictions`
      and `labels` are incompatible.
    """
    if from_logits:
      log_prob = jax.nn.log_softmax(predictions, axis=-1)
    else:
      predictions = base.divide_no_nan(
          predictions, jnp.sum(predictions, axis=-1, keepdims=True)
      )
      epsilon = 1e-7
      predictions = jnp.clip(predictions, epsilon, 1.0 - epsilon)
      log_prob = jnp.log(predictions)

    labels_one_hot = jax.nn.one_hot(labels, predictions.shape[-1], axis=-1)
    crossentropy = -jnp.sum(labels_one_hot * log_prob, axis=-1)

    # Sum across sequence length dimension first.
    if sample_weights is not None:
      crossentropy = crossentropy * sample_weights
      # Normalize by the sum of weights for each sequence.
      crossentropy = base.divide_no_nan(
          jnp.sum(crossentropy), jnp.sum(sample_weights)
      )
    else:
      crossentropy = jnp.mean(crossentropy)

    batch_size = jnp.array(labels.shape[0])
    return cls(
        aggregate_crossentropy=(batch_size * crossentropy),
        num_samples=batch_size,
    )

  def merge(self, other: 'Perplexity') -> 'Perplexity':
    return type(self)(
        aggregate_crossentropy=(
            self.aggregate_crossentropy + other.aggregate_crossentropy
        ),
        num_samples=self.num_samples + other.num_samples,
    )

  def compute(self) -> jax.Array:
    return jnp.exp(
        base.divide_no_nan(self.aggregate_crossentropy, self.num_samples)
    )


@flax.struct.dataclass
class RougeBase(clu_metrics.Metric, abc.ABC):
  """Abstract base class for ROUGE metrics using macro-averaging."""

  total_precision: jax.Array
  total_recall: jax.Array
  total_f1: jax.Array
  num_examples: jax.Array

  @classmethod
  @abc.abstractmethod
  def empty(cls, **kwargs) -> 'RougeBase':
    """Creates an empty Rouge metric. Implemented by subclasses."""
    raise NotImplementedError('Subclasses must implement the empty method.')

  @staticmethod
  @abc.abstractmethod
  def _calculate_instance_scores(
      pred_tokens: list[str], ref_tokens: list[str], **kwargs
  ) -> tuple[float, float, float]:
    """Calculates precision, recall, and F1 for a single prediction-reference pair.

    kwargs may contain 'order' for RougeN.

    Returns:
        A tuple (precision, recall, f1_score).
    """
    raise NotImplementedError(
        'Subclasses must implement _calculate_instance_scores.'
    )

  @classmethod
  def _get_common_initial_values(cls) -> dict[str, jax.Array]:
    """Returns a dictionary of common metric attributes initialized to zero."""
    return {
        'total_precision': jnp.array(0.0, jnp.float32),
        'total_recall': jnp.array(0.0, jnp.float32),
        'total_f1': jnp.array(0.0, jnp.float32),
        'num_examples': jnp.array(0.0, jnp.float32),
    }

  @classmethod
  def _get_specific_constructor_args_for_class(cls) -> dict[str, int]:
    """Returns subclass-specific keyword arguments for the constructor."""
    return {}

  @classmethod
  def from_model_output(
      cls, predictions: list[str], references: list[str], **kwargs
  ) -> 'RougeBase':
    """Computes sums of per-instance ROUGE scores for a batch.

    Subclasses implement _calculate_instance_scores and helper methods for
    specific constructor arguments and validation.

    Args:
      predictions: A list of predicted strings. The shape should be (batch_size,
        ).
      references: A list of reference strings. The shape should be (batch_size,
        ).
      **kwargs: Additional keyword arguments passed to
        _calculate_instance_scores and to the class constructor.

    Returns:
      An instance of the ROUGE metric with accumulated scores.
    """
    total_precision = 0.0
    total_recall = 0.0
    total_f1 = 0.0
    num_examples = 0.0

    for pred_str, ref_str in zip(predictions, references):
      pred_tokens = pred_str.split()
      ref_tokens = ref_str.split()

      precision, recall, f1 = cls._calculate_instance_scores(
          pred_tokens, ref_tokens, **kwargs
      )

      total_precision += precision
      total_recall += recall
      total_f1 += f1
      num_examples += 1

    constructor_args = {
        'total_precision': jnp.array(total_precision, dtype=jnp.float32),
        'total_recall': jnp.array(total_recall, dtype=jnp.float32),
        'total_f1': jnp.array(total_f1, dtype=jnp.float32),
        'num_examples': jnp.array(num_examples, dtype=jnp.float32),
    }
    constructor_args.update(
        cls._get_specific_constructor_args_for_class(**kwargs)
    )
    return cls(**constructor_args)

  def merge(self, other: 'RougeBase') -> 'RougeBase':
    """Merges this Rouge metric with another."""
    if not isinstance(other, type(self)):
      raise TypeError(
          'Cannot merge instances of different types:'
          f' {type(self).__name__} and {type(other).__name__}'
      )

    self._validate_merge_specifics(other)

    merged_data = {
        'total_precision': self.total_precision + other.total_precision,
        'total_recall': self.total_recall + other.total_recall,
        'total_f1': self.total_f1 + other.total_f1,
        'num_examples': self.num_examples + other.num_examples,
    }
    merged_data.update(self._get_specific_fields_for_merge_constructor())
    return type(self)(**merged_data)  # type: ignore[call-arg]

  def _validate_merge_specifics(self, other: 'RougeBase'):
    """Hook for subclass-specific validation during merge (e.g., check order).

    Args:
      other: The other instance of RougeBase to merge with.

    Raises:
      ValueError if validation fails.
    """
    pass

  def _get_specific_fields_for_merge_constructor(self) -> dict[str, int]:
    """Returns subclass-specific fields for constructing a merged instance.

    Returns:
      A dictionary of subclass-specific fields to merge. These fields (e.g.,
      `order` for RougeN) are taken from `self`.
    """
    return {}

  def compute(self) -> jax.Array:
    """Computes macro-averaged recall, precision, and F1-score."""
    macro_avg_precision = base.divide_no_nan(
        self.total_precision, self.num_examples
    )
    macro_avg_recall = base.divide_no_nan(self.total_recall, self.num_examples)
    macro_avg_f1score = base.divide_no_nan(self.total_f1, self.num_examples)

    return jnp.stack([macro_avg_precision, macro_avg_recall, macro_avg_f1score])


@flax.struct.dataclass
class RougeL(RougeBase):
  r"""Computes macro-averaged ROUGE-L recall, precision, and F1-score.

  ROUGE-L measures the longest common subsequence (LCS) between a prediction
  and a reference. This metric calculates ROUGE-L precision, recall, and
  F1-score for each individual prediction compared against its single
  corresponding reference. These per-instance scores are then averaged.

  How ROUGE-L scores are calculated for each individual prediction-reference
  pair:

  For a single prediction P and reference R:

  .. math::
      LCS(P, R) = \text{length of the Longest Common Subsequence}
  .. math::
      \text{Recall}_{\text{LCS}} = \frac{LCS(P, R)}{|R|}
  .. math::
      \text{Precision}_{\text{LCS}} = \frac{LCS(P, R)}{|P|}
  .. math::
      \text{F1}_{\text{LCS}} = 2 \times \frac{\text{Precision} \times
      \text{Recall}}{\text{Precision} + \text{Recall}}

  Final Macro-Averaged Metrics:

  .. math::
      \text{MacroAvgPrecision} =
      \frac{\text{total_precision}}{\text{num_examples}}
  .. math::
      \text{MacroAvgRecall} = \frac{\text{total_recall}}{\text{num_examples}}
  .. math::
      \text{MacroAvgF1} = \frac{\text{total_f1}}{\text{num_examples}}

  Attributes:
    total_precision: Accumulated sum of LCS precision scores from each instance.
    total_recall: Accumulated sum of LCS recall scores from each instance.
    total_f1: Accumulated sum of LCS F1 scores from each instance.
    num_examples: The number of instances (prediction-reference pairs)
      processed.
  """

  @classmethod
  def empty(cls, **kwargs) -> 'RougeL':
    common_values = super()._get_common_initial_values()
    return cls(**common_values)

  @staticmethod
  def _calculate_instance_scores(
      pred_tokens: list[str],
      ref_tokens: list[str],
      **kwargs,
  ) -> tuple[float, float, float]:
    lcs = jnp.array(_lcs_length(pred_tokens, ref_tokens))
    pred_len = jnp.array(len(pred_tokens))
    ref_len = jnp.array(len(ref_tokens))

    precision = base.divide_no_nan(lcs, pred_len).item()
    recall = base.divide_no_nan(lcs, ref_len).item()
    f1 = base.divide_no_nan(2 * precision * recall, precision + recall).item()
    return float(precision), float(recall), float(f1)


@flax.struct.dataclass
class RougeN(RougeBase):
  r"""Computes macro-averaged ROUGE-N recall, precision, and F1-score.

  This metric first calculates ROUGE-N precision, recall, and F1-score for each
  individual prediction compared against its single corresponding reference.
  ROUGE-N scores are based on the number of overlapping n-grams (sequences of n
  words) between the prediction and the reference text. These per-instance
  precision, recall, and F1-scores are then averaged across all instances in
  the dataset/batch.

  How ROUGE-N scores are calculated for each individual prediction-reference
  pair:

  .. math::
      \text{Precision} = \frac{N_o}{N_p}
  .. math::
      \text{Recall} = \frac{N_o}{N_r}
  .. math::
      \text{F1} = 2 \times \frac{\text{Precision} \times
      \text{Recall}}{\text{Precision} + \text{Recall}}
  where:
      - :math:`N_o` be the number of n-grams that overlap between the prediction
      and the reference.
      - :math:`N_p` be the total number of n-grams in the prediction.
      - :math:`N_r` be the total number of n-grams in the reference.

  Final Macro-Averaged Metrics:

  .. math::
      \text{MacroAvgPrecision} =
      \frac{\text{total_precision}}{\text{num_examples}}
  .. math::
      \text{MacroAvgRecall} = \frac{\text{total_recall}}{\text{num_examples}}
  .. math::
      \text{MacroAvgF1} = \frac{\text{total_f1}}{\text{num_examples}}

  Attributes:
    order: The specific 'N' in ROUGE-N (e.g., 1 for ROUGE-1, 2 for ROUGE-2).
    total_precision: Accumulated sum of precision scores from each instance.
    total_recall: Accumulated sum of recall scores from each instance.
    total_f1: Accumulated sum of f1 scores from each instance.
    num_examples: The number of instances (prediction-reference pairs)
      processed.
  """

  order: int

  @classmethod
  def empty(cls, order: int = 2) -> 'RougeN':
    common_values = super()._get_common_initial_values()
    return cls(order=order, **common_values)

  @classmethod
  def _get_specific_constructor_args_for_class(cls, **kwargs) -> dict[str, int]:
    return {'order': kwargs.get('order', 2)}

  @staticmethod
  def _calculate_instance_scores(
      pred_tokens: list[str], ref_tokens: list[str], **kwargs
  ) -> tuple[float, float, float]:
    order = kwargs.get('order', 2)

    pred_ngrams_counts = _get_single_n_grams(pred_tokens, order)
    ref_ngrams_counts = _get_single_n_grams(ref_tokens, order)
    overlap_counts = pred_ngrams_counts & ref_ngrams_counts

    prediction_ngrams = jnp.array(sum(pred_ngrams_counts.values()))
    reference_ngrams = jnp.array(sum(ref_ngrams_counts.values()))
    overlapping_ngrams = jnp.array(sum(overlap_counts.values()))

    precision = base.divide_no_nan(overlapping_ngrams, prediction_ngrams)
    recall = base.divide_no_nan(overlapping_ngrams, reference_ngrams)
    f1 = base.divide_no_nan(2 * precision * recall, precision + recall)
    return float(precision), float(recall), float(f1)

  def _validate_merge_specifics(self, other: 'RougeBase'):
    if not isinstance(other, RougeN) or self.order != other.order:
      raise ValueError(
          'RougeN metrics with different orders cannot be merged. '
          f'Got {self.order} and {getattr(other, "order", "N/A")}.'
      )

  def _get_specific_fields_for_merge_constructor(self) -> dict[str, int]:
    return {'order': self.order}


@flax.struct.dataclass
class WER(base.Average):
  r"""Computes Word Error Rate (WER) for speech recognition or text generation tasks.

  Word Error Rate measures the edit distance between reference texts and
  predictions,
  normalized by the length of the reference texts. It is calculated as:

  .. math::
      WER = \frac{S + D + I}{N}

  where:
      - S is the number of substitutions
      - D is the number of deletions
      - I is the number of insertions
      - N is the number of words in the reference

  A lower WER indicates better performance, with 0 being perfect.

  This implementation accepts both pre-tokenized inputs (lists of tokens) and
  untokenized
  strings. When strings are provided, they are tokenized by splitting on
  whitespace.
  """

  @classmethod
  def from_model_output(
      cls,
      predictions: list[str],
      references: list[str],
  ) -> 'WER':
    """Updates the metric.

    Args:
        prediction: Either a string or a list of tokens in the predicted
          sequence.
        reference: Either a string or a list of tokens in the reference
          sequence.

    Returns:
        New WER metric instance.

    Raises:
        ValueError: If inputs are not properly formatted or are empty.
    """
    if not predictions or not references:
      raise ValueError('predictions and references must not be empty')

    if isinstance(predictions, str):
      predictions = predictions.split()
    if isinstance(references, str):
      references = references.split()

    edit_distance = cls._levenshtein_distance(predictions, references)
    reference_length = len(references)

    return cls(
        total=jnp.array(edit_distance, dtype=jnp.float32),
        count=jnp.array(reference_length, dtype=jnp.float32),
    )

  @staticmethod
  def _levenshtein_distance(prediction: list, reference: list) -> int:
    """Computes the Levenshtein (edit) distance between two token sequences.

    Args:
        prediction: List of tokens in the predicted sequence.
        reference: List of tokens in the reference sequence.

    Returns:
        The minimum number of edits needed to transform prediction into
        reference.
    """
    m, n = len(prediction), len(reference)

    # Handle edge cases
    if m == 0:
      return n
    if n == 0:
      return m

    # Create distance matrix
    distance_matrix = [[0 for _ in range(n + 1)] for _ in range(m + 1)]

    # Initialize first row and column
    for i in range(m + 1):
      distance_matrix[i][0] = i
    for j in range(n + 1):
      distance_matrix[0][j] = j

    # Fill the matrix
    for i in range(1, m + 1):
      for j in range(1, n + 1):
        cost = 0 if prediction[i - 1] == reference[j - 1] else 1

        distance_matrix[i][j] = min(
            distance_matrix[i - 1][j] + 1,  # deletion
            distance_matrix[i][j - 1] + 1,  # insertion
            distance_matrix[i - 1][j - 1] + cost,  # substitution
        )

    return distance_matrix[m][n]
